"""
Base declarative class for all ORM
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import inspect
from sqlalchemy.orm import as_declarative

if TYPE_CHECKING:
    from typing import Any, TypeVar, Type, Dict, Optional, Iterable, Union

    try:
        from pydantic.v1 import BaseModel
    except ImportError:
        from pydantic import BaseModel

    _T = TypeVar("_T")


@as_declarative()
class BaseORM:
    """Base declarative class of all ORM"""

    @classmethod
    def from_model(cls, model_data: Union[dict, BaseModel]):
        """
        Converts a pydantic model or dictionary to this ORM type

        By default, we just construct the ORM from the fields of the model. If they don't match one-to-one,
        then this function should be overridden by the derived class
        """

        if isinstance(model_data, dict):
            return cls(**model_data)
        else:
            return cls(**model_data.dict())

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        """
        Converts the ORM to a dictionary that corresponds to the QCPortal pydantic model

        All columns are included by default, but some can be removed using the exclude parameter.
        """

        d = self.__dict__.copy()
        d.pop("_sa_instance_state")

        # remove typical model excludes
        if hasattr(self, "_qcportal_model_excludes"):
            for k in self._qcportal_model_excludes:
                d.pop(k, None)

        # remove any manually specified excludes
        if exclude is not None:
            for k in exclude:
                d.pop(k, None)

        for k, v in d.items():
            if isinstance(v, BaseORM):
                d[k] = v.model_dict()
            elif isinstance(v, list):
                d[k] = [x.model_dict() if isinstance(x, BaseORM) else x for x in v]
            elif isinstance(v, dict):
                d[k] = {x: y.model_dict() if isinstance(y, BaseORM) else y for x, y in v.items()}

        return d

    def to_model(self, as_type: Type[_T]) -> _T:
        """
        Converts this ORM to a particular type or pydantic model

        This will convert this ORM to a type that has matching columns. For example,
        MoleculeORM to a Molecule

        Parameters
        ----------
        as_type
            Type to convert to

        Returns
        -------
        :
            An object of type Type
        """

        return as_type(**self.model_dict())

    def to_insert_dict(self) -> Dict[str, Any]:
        """
        Converts this model into a dictionary that can be inserted into the database

        This is useful for using SQLAlchemy constructs that take dictionaries rather than ORM objects.

        - Skips autoincrement PKs if None
        - Skips computed / server-generated columns
        """

        state = inspect(self)
        mapper = state.mapper
        table = mapper.local_table

        d = {}
        for col in table.columns:
            # Remove any autoincrement primary keys that are manually set to None. If set, then
            # it will try to be inserted into the db, resulting in a not-null violation
            if col.primary_key and col.autoincrement and getattr(self, col.name, None) is None:
                continue

            if col.computed:
                continue

            if hasattr(self, col.name):
                d[col] = getattr(self, col.name)

        return d



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import Column, Integer, String, JSON, Float, Index, CHAR, UniqueConstraint
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import column_property

from qcfractal.db_socket.base_orm import BaseORM
from qcfractal.db_socket.column_types import MsgpackExt
from qcportal.molecules import Molecule

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class MoleculeORM(BaseORM):
    """
    Table for storing molecules
    """

    __tablename__ = "molecule"

    id = Column(Integer, primary_key=True)

    molecule_hash = Column(CHAR(40))  # sha1 is always 40 chars

    # Required data
    schema_name = Column(String)
    schema_version = Column(Integer, default=2)
    symbols = Column(MsgpackExt, nullable=False)
    geometry = Column(MsgpackExt, nullable=False)

    # Molecule data
    name = Column(String, default="")
    identifiers = Column(JSONB)
    comment = Column(String)
    molecular_charge = Column(Float, default=0)
    molecular_multiplicity = Column(Integer, default=1)

    # Atom data
    masses = Column(MsgpackExt)
    real = Column(MsgpackExt)
    atom_labels = Column(MsgpackExt)
    atomic_numbers = Column(MsgpackExt)
    mass_numbers = Column(MsgpackExt)

    # Fragment and connection data
    connectivity = Column(JSON)
    fragments = Column(MsgpackExt)
    fragment_charges = Column(JSON)  # Column(ARRAY(Float))
    fragment_multiplicities = Column(JSON)  # Column(ARRAY(Integer))

    # Orientation & symmetry
    fix_symmetry = Column(String)

    # These are always forced to be true
    fix_com = column_property(True)
    fix_orientation = column_property(True)

    # Molecule is always validated before going in the database
    validated = column_property(True)

    # Extra
    provenance = Column(JSON)
    extras = Column(JSON)

    __table_args__ = (
        Index("ix_molecule_identifiers", "identifiers", postgresql_using="gin"),
        UniqueConstraint("molecule_hash", name="ux_molecule_molecule_hash"),
    )

    _qcportal_model_excludes = ["molecule_hash"]

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)

        # TODO - this is because the pydantic models are goofy
        return {k: v for k, v in d.items() if v is not None}

    @classmethod
    def from_model(cls, model_data: dict | Molecule):
        """
        Convert a pydantic (QCElemental) Molecule to an ORM
        """

        # Validate the molecule if it hasn't been validated already
        if isinstance(model_data, dict):
            if not model_data["validated"]:
                molecule = Molecule(**model_data, validate=True)
            else:
                molecule = Molecule(**model_data)
        else:
            if not model_data.validated:
                molecule = Molecule(**model_data.dict())
            else:
                molecule = model_data

        mol_dict = molecule.dict(exclude={"id", "validated", "fix_com", "fix_orientation"})

        # Build these quantities fresh from what is actually stored
        mol_dict["molecule_hash"] = model_data.get_hash()

        mol_dict.setdefault("identifiers", dict())
        mol_dict["identifiers"]["molecule_hash"] = mol_dict["molecule_hash"]
        mol_dict["identifiers"]["molecular_formula"] = molecule.get_molecular_formula()

        return cls(**mol_dict)



"""rename owner_user, remove owner_group

Revision ID: c5a3bed43646
Revises: db730926b1b9
Create Date: 2025-04-11 13:04:03.146646

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "c5a3bed43646"
down_revision = "db730926b1b9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index("ix_base_dataset_owner_group_id", table_name="base_dataset")
    op.drop_index("ix_base_dataset_owner_user_id", table_name="base_dataset")
    op.drop_constraint("base_dataset_owner_user_id_fkey", "base_dataset", type_="foreignkey")
    op.drop_constraint("base_dataset_owner_user_id_owner_group_id_fkey", "base_dataset", type_="foreignkey")
    op.drop_constraint("base_dataset_owner_group_id_fkey", "base_dataset", type_="foreignkey")
    op.alter_column("base_dataset", "owner_user_id", new_column_name="creator_user_id")
    op.drop_column("base_dataset", "owner_group_id")
    op.create_index("ix_base_dataset_creator_user_id", "base_dataset", ["creator_user_id"], unique=False)
    op.create_foreign_key(None, "base_dataset", "user", ["creator_user_id"], ["id"])

    op.drop_index("ix_base_record_owner_group_id", table_name="base_record")
    op.drop_index("ix_base_record_owner_user_id", table_name="base_record")
    op.drop_constraint("base_record_owner_group_id_fkey", "base_record", type_="foreignkey")
    op.drop_constraint("base_record_owner_user_id_fkey", "base_record", type_="foreignkey")
    op.drop_constraint("base_record_owner_user_id_owner_group_id_fkey", "base_record", type_="foreignkey")
    op.drop_column("base_record", "owner_group_id")
    op.alter_column("base_record", "owner_user_id", new_column_name="creator_user_id")
    op.create_index("ix_base_record_creator_user_id", "base_record", ["creator_user_id"], unique=False)
    op.create_foreign_key(None, "base_record", "user", ["creator_user_id"], ["id"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise NotImplementedError("Downgrade is not supported")
    # ### end Alembic commands ###



import pytest

from qcarchivetesting import test_users, test_groups
from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcportal import PortalRequestError
from qcportal.auth import GroupInfo
from qcportal.exceptions import (
    InvalidGroupnameError,
)
from .test_group_socket import invalid_groupnames


def test_group_client_list(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])
    groups = client.list_groups()

    assert len(groups) == 4

    for g in groups:
        assert g.groupname in test_groups


def test_group_client_get(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    for groupname in test_groups:
        g = client.get_group(groupname)
        assert g.groupname == groupname

        g2 = client.get_group(g.id)
        assert g == g2


def test_group_client_add(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    client.add_group(GroupInfo(groupname="a_group", description="A description"))
    g = client.get_group("a_group")
    assert g.groupname == "a_group"
    assert g.id is not None
    assert g.description == "A description"


def test_group_client_add_with_id(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    with pytest.raises(RuntimeError, match=r"Cannot add group.*contains an id"):
        ginfo = GroupInfo(id=1234, groupname="a_group", description="A description")
        client.add_group(ginfo)


def test_group_client_add_existing(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    with pytest.raises(PortalRequestError, match=r"already exists"):
        ginfo = GroupInfo(groupname="group1", description="A description")
        client.add_group(ginfo)


def test_group_client_use_nonexist(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    with pytest.raises(PortalRequestError, match=r"Group.*does not exist"):
        client.get_group("no_group")
    with pytest.raises(PortalRequestError, match=r"Group.*does not exist"):
        client.get_group(1234)
    with pytest.raises(PortalRequestError, match=r"Group.*does not exist"):
        client.delete_group("no_group")
    with pytest.raises(PortalRequestError, match=r"Group.*does not exist"):
        client.delete_group(1234)


def test_group_client_use_invalid_groupname(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    for groupname in invalid_groupnames:
        with pytest.raises(InvalidGroupnameError):
            client.get_group(groupname)
        with pytest.raises(InvalidGroupnameError):
            client.delete_group(groupname)


def test_group_client_delete(secure_snowflake: QCATestingSnowflake):
    client = secure_snowflake.client("admin_user", test_users["admin_user"]["pw"])

    user = client.get_user("admin_user")
    assert "group1" in user.groups
    assert "group2" in user.groups

    gid2 = client.get_group("group2").id

    client.delete_group("group1")
    client.delete_group(gid2)

    with pytest.raises(PortalRequestError, match=r"Group.*does not exist"):
        client.get_group("group1")
    with pytest.raises(PortalRequestError, match=r"Group.*does not exist"):
        client.get_group(gid2)

    user = client.get_user("admin_user")
    assert "group1" not in user.groups
    assert "group2" not in user.groups



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import pytest

from qcarchivetesting import load_molecule_data
from qcportal import PortalRequestError
from qcportal.molecules import Molecule, MoleculeIdentifiers

if TYPE_CHECKING:
    from qcportal import PortalClient


def test_molecules_client_basic(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")

    # Add once
    meta, ids = snowflake_client.add_molecules([water])
    assert meta.success
    assert meta.n_inserted == 1
    assert meta.n_existing == 0
    assert meta.inserted_idx == [0]

    # Now get the molecule by id
    mols = snowflake_client.get_molecules(ids)
    assert len(mols) == 1
    assert mols[0].get_hash() == water.get_hash()
    assert water == mols[0]

    # Get as a single id
    mol = snowflake_client.get_molecules(ids[0])
    assert isinstance(mol, Molecule)
    assert water == mol

    # Is a valid molecule object dictionary
    assert mols[0].validated


def test_molecules_client_get_nonexist(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")
    meta, ids = snowflake_client.add_molecules([water])

    mols = snowflake_client.get_molecules(123456, missing_ok=True)
    assert mols is None

    mols = snowflake_client.get_molecules([123, 456, 789, ids[0]], missing_ok=True)

    assert len(mols) == 4
    assert mols[0] is None
    assert mols[1] is None
    assert mols[2] is None
    assert water == mols[3]

    # Now try with missing_ok = False. This should raise an exception
    with pytest.raises(PortalRequestError, match=r"Could not find all requested records"):
        snowflake_client.get_molecules(123, missing_ok=False)

    with pytest.raises(PortalRequestError, match=r"Could not find all requested records"):
        snowflake_client.get_molecules([123, 456, 789, ids[0]], missing_ok=False)


def test_molecules_client_add_with_id(snowflake_client: PortalClient):
    # Adding with an id already set is ok - the returned id may not
    # be the same

    bad_id = 99998888
    water = load_molecule_data("water_dimer_minima")
    water = water.copy(update={"id": bad_id})

    meta, ids = snowflake_client.add_molecules([water])
    assert meta.success
    assert meta.n_inserted == 1
    assert meta.n_existing == 0
    assert meta.inserted_idx == [0]
    assert ids[0] != water.id

    ## Getting by the old id shouldn't work
    with pytest.raises(PortalRequestError, match=r"Could not find all requested records"):
        snowflake_client.get_molecules([bad_id], missing_ok=False)

    ## Getting by the new id should work
    mols = snowflake_client.get_molecules(ids, missing_ok=False)
    assert mols[0].id == ids[0]
    assert mols[0] == water


def test_molecules_client_add_duplicates(snowflake_client: PortalClient):
    # Tests various ways of adding duplicate molecules
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")

    # Duplicate molecules in same add_molecules call
    meta, ids = snowflake_client.add_molecules([water, hooh, water, hooh, hooh])
    assert meta.success
    assert meta.n_inserted == 2
    assert meta.n_existing == 3
    assert meta.inserted_idx == [0, 1]
    assert meta.existing_idx == [2, 3, 4]
    assert ids[0] == ids[2]
    assert ids[1] == ids[3]
    assert ids[1] == ids[4]

    # Test in a different order
    meta, ids2 = snowflake_client.add_molecules([hooh, hooh, water, water, hooh])
    assert meta.success
    assert meta.n_inserted == 0
    assert meta.n_existing == 5
    assert meta.inserted_idx == []
    assert meta.existing_idx == [0, 1, 2, 3, 4]
    assert ids2[0] == ids2[1]
    assert ids2[0] == ids2[4]
    assert ids2[2] == ids2[3]

    assert ids2[0] == ids[1]
    assert ids2[2] == ids[0]


def test_molecules_client_get_duplicates(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")

    meta, ids = snowflake_client.add_molecules([water, hooh])

    id1, id2 = ids
    mols = snowflake_client.get_molecules([id1, id2, id1, id2, id1])
    assert len(mols) == 5
    assert mols[0] == mols[2]
    assert mols[0] == mols[4]
    assert mols[1] == mols[3]
    assert mols[0].id == mols[2].id
    assert mols[0].id == mols[4].id
    assert mols[1].id == mols[3].id

    # Try getting in a different order
    mols = snowflake_client.get_molecules([id2, id1, id1, id1, id2])
    assert len(mols) == 5
    assert mols[0] == mols[4]
    assert mols[1] == mols[2]
    assert mols[1] == mols[3]
    assert mols[0].id == mols[4].id
    assert mols[1].id == mols[2].id
    assert mols[1].id == mols[3].id


def test_molecules_client_get_empty(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")
    _, ids = snowflake_client.add_molecules([water])

    assert snowflake_client.get_molecules([]) == []


def test_molecules_client_delete(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")
    meta, ids = snowflake_client.add_molecules([water])

    meta = snowflake_client.delete_molecules(ids)
    assert meta.success is True
    assert meta.n_deleted == 1
    assert meta.n_errors == 0
    assert meta.deleted_idx == [0]

    mol = snowflake_client.get_molecules(ids[0], missing_ok=True)
    assert mol is None


def test_molecules_client_delete_nonexist(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")
    meta, ids = snowflake_client.add_molecules([water])
    assert meta.success

    # Deletion succeeds if it doesn't exist
    meta = snowflake_client.delete_molecules([456, ids[0], ids[0], 123, 789])
    assert meta.success is True
    assert meta.n_deleted == 5
    assert meta.n_errors == 0
    assert meta.deleted_idx == [0, 1, 2, 3, 4]


def test_molecules_client_delete_inuse(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")

    meta, water_ids = snowflake_client.add_molecules([water])
    snowflake_client.add_singlepoints(water, "prog1", "energy", "b3lyp", "sto-3g")

    meta = snowflake_client.delete_molecules(water_ids)
    assert meta.success is False
    assert meta.n_deleted == 0
    assert meta.n_errors == 1
    assert meta.error_idx == [0]
    assert "may still be referenced" in meta.errors[0][1]


def test_molecules_client_modify(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")
    hooh = load_molecule_data("hooh")

    _, ids = snowflake_client.add_molecules([water, hooh])
    assert len(ids) == 2

    meta = snowflake_client.modify_molecule(
        ids[0],
        name="water_dimer",
        comment="This is a comment",
        identifiers=MoleculeIdentifiers(smiles="madeupsmiles", molecule_hash="notahash", molecular_formula="XXXX"),
        overwrite_identifiers=False,
    )
    assert meta.success
    assert meta.updated_idx == [0]

    # Did it actually update?
    mols = snowflake_client.get_molecules(ids)
    assert mols[0].name == "water_dimer"
    assert mols[0].comment == "This is a comment"
    assert mols[0].identifiers.smiles == "madeupsmiles"
    assert mols[0].identifiers.inchi is None

    # Hash & formula unchanged
    assert mols[0].identifiers.molecule_hash == water.get_hash()
    assert mols[0].identifiers.molecular_formula == water.get_molecular_formula()

    # Update again
    meta = snowflake_client.modify_molecule(
        ids[0],
        name="water_dimer",
        comment="This is a comment",
        identifiers=MoleculeIdentifiers(
            inchi="madeupinchi",
        ),
        overwrite_identifiers=False,
    )
    assert meta.success
    assert meta.updated_idx == [0]

    # Updated, but did not overwrite
    mols = snowflake_client.get_molecules(ids)
    assert mols[0].name == "water_dimer"
    assert mols[0].comment == "This is a comment"
    assert mols[0].identifiers.smiles == "madeupsmiles"
    assert mols[0].identifiers.inchi == "madeupinchi"  # newly added

    # Hash & formula unchanged
    assert mols[0].identifiers.molecule_hash == water.get_hash()
    assert mols[0].identifiers.molecular_formula == water.get_molecular_formula()

    # Now try with overwrite_identifiers = True
    meta = snowflake_client.modify_molecule(
        ids[0],
        name="water_dimer 2",
        identifiers=MoleculeIdentifiers(inchikey="madeupinchikey"),
        overwrite_identifiers=True,
    )
    assert meta.success
    assert meta.updated_idx == [0]

    # Did it actually update?
    mols = snowflake_client.get_molecules(ids)
    assert mols[0].name == "water_dimer 2"
    assert mols[0].comment == "This is a comment"
    assert mols[0].identifiers.smiles is None
    assert mols[0].identifiers.inchi is None
    assert mols[0].identifiers.inchikey == "madeupinchikey"

    # Hash & formula unchanged
    assert mols[0].identifiers.molecule_hash == water.get_hash()
    assert mols[0].identifiers.molecular_formula == water.get_molecular_formula()


def test_molecules_client_update_nonexist(snowflake_client: PortalClient):
    water = load_molecule_data("water_dimer_minima")

    _, ids = snowflake_client.add_molecules([water])
    assert len(ids) == 1

    with pytest.raises(PortalRequestError, match=r"Molecule with id.*not found in the database"):
        snowflake_client.modify_molecule(
            9999,
            name="water_dimer",
            comment="This is a comment",
            identifiers=MoleculeIdentifiers(smiles="madeupsmiles", inchi="madeupinchi"),
            overwrite_identifiers=False,
        )


def test_molecules_client_upload_files(snowflake_client: PortalClient):
    from qcarchivetesting.helpers import _my_path

    data_path = os.path.join(_my_path, "molecule_data")

    # These are the hashes for the various files
    file_hashes = {
        "water_stacked.json": "27061c76f7de3eca6a2d9521d660661fe177192f",
        "benzene_dimer.xyz": "2a7b21a54fb2adf8ef4064b8aa25804a24a61781",
        "benzene_dimer.json": "0990d963680c22adde43bd650dff75e3b16808be",
    }

    file_paths = [
        os.path.join(data_path, "test_archive_1.zip"),
        os.path.join(data_path, "test_archive_1.tar.gz"),
        os.path.join(data_path, "test_archive_1.tar.bz2"),
        os.path.join(data_path, "test_archive_1.tar.xz"),
        os.path.join(data_path, "benzene_dimer.xyz"),
        os.path.join(data_path, "water_stacked.json"),
    ]

    assert all(os.path.isfile(x) for x in file_paths)

    minfo, errors = snowflake_client.upload_molecules(file_paths)
    assert len(errors) == 0

    for k, v in minfo.items():
        for fname, mid in v:
            mol = snowflake_client.get_molecules(mid)
            assert mol.get_hash() == file_hashes[fname]



from __future__ import annotations

import contextlib
import importlib
import io
import logging
import textwrap
from typing import List, Dict, Tuple, Optional, Sequence, Any, Union, TYPE_CHECKING

import tabulate
from sqlalchemy import select, func
from sqlalchemy.dialects.postgresql import array_agg, aggregate_order_by
from sqlalchemy.orm import defer, undefer, lazyload, joinedload, selectinload

from qcfractal.components.services.db_models import ServiceQueueORM, ServiceDependencyORM
from qcfractal.components.singlepoint.record_db_models import QCSpecificationORM
from qcfractal.db_socket.helpers import insert_general
from qcportal.exceptions import MissingDataError
from qcportal.manybody import (
    ManybodySpecification,
    ManybodyQueryFilters,
    ManybodyInput,
    ManybodyMultiInput,
    ManybodyRecord,
)
from qcportal.metadata_models import InsertMetadata, InsertCountsMetadata
from qcportal.molecules import Molecule
from qcportal.record_models import PriorityEnum, RecordStatusEnum, OutputTypeEnum
from qcportal.utils import chunk_iterable, hash_dict, is_included
from .record_db_models import (
    ManybodyClusterORM,
    ManybodyRecordORM,
    ManybodySpecificationORM,
    ManybodySpecificationLevelsORM,
)
from ..base_record_socket import BaseRecordSocket
from ..record_utils import append_output

_qcm_spec = importlib.util.find_spec("qcmanybody")

if _qcm_spec is not None:
    qcmanybody = importlib.util.module_from_spec(_qcm_spec)
    _qcm_spec.loader.exec_module(qcmanybody)

if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.db_socket.socket import SQLAlchemySocket


# Meaningless, but unique to manybody
manybody_insert_lock_id = 14500
manybody_insert_spec_lock_id = 14501


def _get_qcmanybody_core(
    mb_orm: ManybodyRecordORM,
) -> Tuple[qcmanybody.ManyBodyCore, Dict[str, ManybodySpecificationLevelsORM]]:
    init_mol: Molecule = mb_orm.initial_molecule.to_model(Molecule)

    qcm_levels = {}
    level_spec_map = {}
    sp_id_map = {}

    for nb, lvl in sorted(mb_orm.specification.levels.items()):
        if nb == -1:
            nb = "supersystem"

        sp_spec = lvl.singlepoint_specification
        if sp_spec.id in sp_id_map:
            sp_name = sp_id_map[sp_spec.id]
        else:
            test_name = f"{sp_spec.program}/{sp_spec.method}/{sp_spec.basis}"
            sp_name = test_name

            # duplicates
            i = 0
            while sp_name in qcm_levels:
                i += 1
                sp_name = f"{test_name}_{i}"

            sp_id_map[sp_spec.id] = sp_name

        qcm_levels[nb] = sp_name
        level_spec_map[sp_name] = lvl

    qcm = qcmanybody.ManyBodyCore(
        molecule=init_mol,
        levels=qcm_levels,
        bsse_type=[qcmanybody.BsseEnum[x] for x in mb_orm.specification.bsse_correction],
        return_total_data=mb_orm.specification.keywords.get("return_total_data", False),
        supersystem_ie_only=mb_orm.specification.keywords.get("supersystem_ie_only", False),
        embedding_charges=None,
    )

    return qcm, level_spec_map


class ManybodyRecordSocket(BaseRecordSocket):
    """
    Socket for handling manybody computations
    """

    # Used by the base class
    record_orm = ManybodyRecordORM
    record_input_type = ManybodyInput
    record_multi_input_type = ManybodyMultiInput

    def __init__(self, root_socket: SQLAlchemySocket):
        BaseRecordSocket.__init__(self, root_socket)
        self._logger = logging.getLogger(__name__)

    def available(self) -> bool:
        return _qcm_spec is not None

    def initialize_service(self, session: Session, service_orm: ServiceQueueORM) -> None:
        mb_orm: ManybodyRecordORM = service_orm.record

        output = "\n\nCreated manybody calculation\n"
        output += "qcmanybody version: " + qcmanybody.__version__ + "\n\n"

        output += "-" * 80 + "\nSpecification:\n\n"

        table_rows = []

        for k, v in mb_orm.specification.keywords.items():
            table_rows.append((k, v))

        if mb_orm.specification.program != "qcmanybody":
            raise RuntimeError(f"Unknown program: {mb_orm.specification.program}")

        table_rows.append(("bsse_correction", str(mb_orm.specification.bsse_correction)))
        output += tabulate.tabulate(table_rows, tablefmt="plain")
        output += "\n\n"

        init_mol: Molecule = mb_orm.initial_molecule.to_model(Molecule)
        output += f"Initial molecule: formula={init_mol.get_molecular_formula()} id={mb_orm.initial_molecule_id}\n"
        output += f"Initial molecule has {len(init_mol.fragments)} fragments\n"

        # Create a computer instance to get what calculations we need
        qcm, spec_map = _get_qcmanybody_core(mb_orm)

        output += "\n\n" + "-" * 80 + "\nModel Chemistries/Specifications:\n\n"
        for name, lvl_spec in spec_map.items():
            output += f"{name}:\n"
            output += textwrap.indent(
                tabulate.tabulate(lvl_spec.singlepoint_specification.model_dict().items(), tablefmt="plain"), "    "
            )
            output += "\n"

        output += "\n\n" + "-" * 80 + "\nLevels:\n\n"
        for level, mc_name in qcm.levels.items():
            output += f" {level:>13}: {mc_name}\n"
        output += "\n\n"

        output += "\n\n" + "-" * 80 + "\nComputation count:\n\n"
        table_rows = []
        for mc, compute_dict in qcm.compute_map.items():
            for nb, frags in compute_dict["all"].items():
                table_rows.append((f"{mc} {nb}-mer", len(frags)))
        output += tabulate.tabulate(table_rows, headers=["n-body", "count"])

        # Add what we need to compute to the database
        table_rows = []

        for mol_batch in chunk_iterable(qcm.iterate_molecules(), 400):
            to_add = [x[2] for x in mol_batch]
            meta, mol_ids = self.root_socket.molecules.add(to_add, session=session)
            if not meta.success:
                raise RuntimeError("Unable to add molecules to the database: " + meta.error_string)

            for (mc_level, label, molecule), mol_id in zip(mol_batch, mol_ids):
                # Decode the label given by qcmanybody
                _, frag, bas = qcmanybody.delabeler(label)

                mb_cluster_orm = ManybodyClusterORM(
                    mc_level=mc_level,
                    fragments=frag,
                    basis=bas,
                    molecule_id=mol_id,
                )
                table_rows.append((mc_level, frag, bas, mol_id, molecule.get_molecular_formula(), molecule.get_hash()))
                mb_orm.clusters.append(mb_cluster_orm)

        output += "\n\nMolecules to compute\n\n"
        output += tabulate.tabulate(
            table_rows, headers=["model chemistry", "fragments", "basis", "molecule_id", "formula", "hash"]
        )

        append_output(session, mb_orm, OutputTypeEnum.stdout, output)

    def iterate_service(
        self,
        session: Session,
        service_orm: ServiceQueueORM,
    ):
        mb_orm: ManybodyRecordORM = service_orm.record

        # Always update with the current provenance
        mb_orm.compute_history[-1].provenance = {
            "creator": "qcmanybody",
            "version": qcmanybody.__version__,
            "routine": "qcmanybody",
        }

        service_orm.dependencies = []

        submitted = []

        qcm, spec_map = _get_qcmanybody_core(mb_orm)
        done_sp_ids = set(c.singlepoint_id for c in mb_orm.clusters if c.singlepoint_id is not None)

        # what we need to submit, mapped by single spec id

        clusters_to_submit = {}
        for c in mb_orm.clusters:
            if c.singlepoint_id is not None:
                continue
            clusters_to_submit.setdefault(c.mc_level, [])
            clusters_to_submit[c.mc_level].append(c)

        for mc_level, clusters in clusters_to_submit.items():
            mol_ids = [c.molecule_id for c in clusters]
            meta, sp_ids = self.root_socket.records.singlepoint.add_internal(
                mol_ids,
                spec_map[mc_level].singlepoint_specification_id,
                service_orm.compute_tag,
                service_orm.compute_priority,
                mb_orm.creator_user_id,
                service_orm.find_existing,
                session=session,
            )

            for cluster, sp_id in zip(clusters, sp_ids):
                cluster.singlepoint_id = sp_id
                submitted.append((cluster, sp_id))

                # Add as a dependency to the service, but only if it's not done yet
                if sp_id not in done_sp_ids:
                    svc_dep = ServiceDependencyORM(record_id=sp_id, extras={})
                    service_orm.dependencies.append(svc_dep)
                    done_sp_ids.add(sp_id)

        if len(submitted) != 0:
            output = f"\nSubmitted {len(submitted)} singlepoint calculations "
            append_output(session, mb_orm, OutputTypeEnum.stdout, output)
            return False

        output = "\n\n" + "*" * 80 + "\n"
        output += "All manybody singlepoint computations are complete!\n\n"

        output += "=" * 20 + "\nSinglepoint results\n" + "=" * 20 + "\n\n"

        # Make a nice output table
        table_rows = []
        for cluster in mb_orm.clusters:
            mol_id = cluster.molecule_id

            energy = cluster.singlepoint_record.properties["return_energy"]
            table_row = [cluster.mc_level, cluster.fragments, cluster.basis, energy, mol_id, cluster.singlepoint_id]
            table_rows.append(table_row)

        output += tabulate.tabulate(
            table_rows,
            headers=["model chemistry", "fragments", "basis", "energy (hartree)", "molecule id", "singlepoint id"],
            floatfmt=".10f",
        )

        # Analyze the actual results
        component_results = {}
        for cluster in mb_orm.clusters:
            mc_level = cluster.mc_level
            label = qcmanybody.labeler(mc_level, cluster.fragments, cluster.basis)
            energy = cluster.singlepoint_record.properties["return_energy"]

            component_results.setdefault(label, {})
            component_results[label]["energy"] = energy

        # Swallow any output
        qcmb_stdout = io.StringIO()

        with contextlib.redirect_stdout(qcmb_stdout):
            mb_orm.properties = qcm.analyze(component_results)

        output += "\n\n" + "=" * 40 + "\nManybody expansion results\n" + "=" * 40 + "\n"
        output += mb_orm.properties.pop("stdout")
        append_output(session, mb_orm, OutputTypeEnum.stdout, output)

        # We are done!
        return True

    def add_specifications(
        self, mb_specs: Sequence[ManybodySpecification], *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, List[int]]:
        """
        Adds a specification for a manybody service to the database, returning its id.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Parameters
        ----------
        mb_specs
            Sequence of specifications to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the IDs of the specification.
        """

        # Because of how we handle levels, we do this the opposite of other record types - we add one at a time

        all_metadata = []
        all_ids = []

        with self.root_socket.optional_session(session) as session:
            for mb_spec in mb_specs:
                meta, spec_id = self.add_specification(mb_spec, session=session)

                if not meta.success:
                    return (
                        InsertMetadata(error_description="Unable to add manybody specification: " + meta.error_string),
                        [],
                    )

                all_metadata.append(meta)
                all_ids.append(spec_id)

        return InsertMetadata.merge(all_metadata), all_ids

    def insert_full_qcportal_records_v1(
        self, session: Session, records: Sequence[ManybodyRecord], creator_user_id: Optional[int]
    ) -> List[ManybodyRecordORM]:
        ret = []

        initial_mols = []
        mb_specs = []

        for record in records:
            initial_mols.append(record.initial_molecule)
            mb_specs.append(record.specification)

        meta, spec_ids = self.root_socket.records.manybody.add_specifications(mb_specs, session=session)
        if not meta.success:
            raise RuntimeError("Aborted manybody insertion - could not add specifications: " + meta.error_string)

        meta, initial_mol_ids = self.root_socket.molecules.add(initial_mols, session=session)
        if not meta.success:
            raise RuntimeError("Aborted manybody insertion - could not add initial molecules: " + meta.error_string)

        for record, initial_mol_id, spec_id in zip(records, initial_mol_ids, spec_ids):
            record_orm = ManybodyRecordORM(
                specification_id=spec_id,
                initial_molecule_id=initial_mol_id,
                status=RecordStatusEnum.complete,
            )

            if record.cluster_records_:
                cluster_mols = [x.molecule for x in record.cluster_records_]
                cluster_records = [x.singlepoint_record for x in record.cluster_records_]

                meta, cluster_mol_ids = self.root_socket.molecules.add(cluster_mols, session=session)
                if not meta.success:
                    raise RuntimeError(
                        "Aborted manybody insertion - could not add initial molecules: " + meta.error_string
                    )

                cluster_sp_ids = self.root_socket.records.insert_full_qcportal_records(
                    session, cluster_records, creator_user_id
                )
                record_orm.clusters = [
                    ManybodyClusterORM(
                        molecule_id=mid,
                        singlepoint_id=spid,
                        mc_level=crec.mc_level,
                        fragments=crec.fragments,
                        basis=crec.basis,
                    )
                    for mid, spid, crec in zip(cluster_mol_ids, cluster_sp_ids, record.cluster_records_)
                ]

            ret.append(record_orm)

        return ret

    def add_specification(
        self, mb_spec: ManybodySpecification, *, session: Optional[Session] = None
    ) -> Tuple[InsertMetadata, Optional[int]]:
        """
        Adds a specification for a manybody service to the database, returning its id.

        If an identical specification exists, then no insertion takes place and the id of the existing
        specification is returned.

        Parameters
        ----------
        mb_spec
            Specification to add to the database
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and the id of the specification.
        """

        mb_kw_dict = mb_spec.keywords.dict()

        mb_spec_dict = {
            "program": mb_spec.program,
            "bsse_correction": sorted(mb_spec.bsse_correction),
            "keywords": mb_kw_dict,
            "protocols": {},
        }
        mb_spec_hash = hash_dict(mb_spec_dict)

        # Map 'supersystem' to -1
        # The reverse (mapping -1 to 'supersystem') happens in the specification orm model_dict function
        levels = mb_spec.levels.copy()
        if "supersystem" in levels:
            levels[-1] = levels.pop("supersystem")

        with self.root_socket.optional_session(session) as session:
            # add all singlepoint specifications

            # Level to singlepoint spec id
            level_spec_id_map: Dict[int, int] = {}
            for k, v in levels.items():
                meta, sp_spec_id = self.root_socket.records.singlepoint.add_specification(qc_spec=v, session=session)

                if not meta.success:
                    return (
                        InsertMetadata(
                            error_description="Unable to add singlepoint specification: " + meta.error_string,
                        ),
                        None,
                    )

                level_spec_id_map[k] = sp_spec_id

            # Now the full manybody specification. Lock due to query + insert
            session.execute(select(func.pg_advisory_xact_lock(manybody_insert_spec_lock_id))).scalar()

            # Create a cte with the specification + levels
            mb_spec_cte = (
                select(
                    ManybodySpecificationORM.id,
                    ManybodySpecificationORM.specification_hash,
                    array_agg(
                        aggregate_order_by(
                            ManybodySpecificationLevelsORM.singlepoint_specification_id,
                            ManybodySpecificationLevelsORM.singlepoint_specification_id.asc(),
                        )
                    ).label("singlepoint_ids"),
                    array_agg(
                        aggregate_order_by(
                            ManybodySpecificationLevelsORM.level,
                            ManybodySpecificationLevelsORM.level.asc(),
                        )
                    ).label("levels"),
                )
                .join(
                    ManybodySpecificationLevelsORM,
                    ManybodySpecificationLevelsORM.manybody_specification_id == ManybodySpecificationORM.id,
                )
                .group_by(ManybodySpecificationORM.id)
                .cte()
            )

            stmt = select(mb_spec_cte.c.id)
            stmt = stmt.where(mb_spec_cte.c.specification_hash == mb_spec_hash)
            stmt = stmt.where(mb_spec_cte.c.levels == sorted(level_spec_id_map.keys()))
            stmt = stmt.where(mb_spec_cte.c.singlepoint_ids == sorted(level_spec_id_map.values()))

            existing_id = session.execute(stmt).scalar_one_or_none()

            if existing_id is not None:
                return InsertMetadata(existing_idx=[0]), existing_id

            # Does not exist. Insert new
            mb_levels_orms = {}
            for level, sp_spec_id in level_spec_id_map.items():
                mb_levels_orms[level] = ManybodySpecificationLevelsORM(
                    level=level, singlepoint_specification_id=sp_spec_id
                )

            new_orm = ManybodySpecificationORM(
                program=mb_spec.program,
                bsse_correction=mb_spec.bsse_correction,
                keywords=mb_kw_dict,
                specification_hash=mb_spec_hash,
                levels=mb_levels_orms,
                protocols={},
            )

            session.add(new_orm)
            session.flush()
            return InsertMetadata(inserted_idx=[0]), new_orm.id

    def get(
        self,
        record_ids: Sequence[int],
        include: Optional[Sequence[str]] = None,
        exclude: Optional[Sequence[str]] = None,
        missing_ok: bool = False,
        *,
        session: Optional[Session] = None,
    ) -> List[Optional[Dict[str, Any]]]:
        options = []
        if include:
            if is_included("initial_molecule", include, exclude, False) or is_included(
                "molecules", include, exclude, False
            ):
                options.append(joinedload(ManybodyRecordORM.initial_molecule))
            if is_included("clusters", include, exclude, False):
                options.append(selectinload(ManybodyRecordORM.clusters).selectinload(ManybodyClusterORM.molecule))

        with self.root_socket.optional_session(session, True) as session:
            return self.root_socket.records.get_base(
                orm_type=self.record_orm,
                record_ids=record_ids,
                include=include,
                exclude=exclude,
                missing_ok=missing_ok,
                additional_options=options,
                session=session,
            )

    def query(
        self,
        query_data: ManybodyQueryFilters,
        *,
        session: Optional[Session] = None,
    ) -> List[int]:
        and_query = []
        need_spec_join = False
        need_qcspec_join = False

        if query_data.program is not None:
            and_query.append(ManybodySpecificationORM.program.in_(query_data.program))
            need_spec_join = True
        if query_data.qc_program is not None:
            and_query.append(QCSpecificationORM.program.in_(query_data.qc_program))
            need_qcspec_join = True
        if query_data.qc_method is not None:
            and_query.append(QCSpecificationORM.method.in_(query_data.qc_method))
            need_qcspec_join = True
        if query_data.qc_basis is not None:
            and_query.append(QCSpecificationORM.basis.in_(query_data.qc_basis))
            need_qcspec_join = True
        if query_data.initial_molecule_id is not None:
            and_query.append(ManybodyRecordORM.initial_molecule_id.in_(query_data.initial_molecule_id))

        stmt = select(ManybodyRecordORM.id)

        if need_spec_join or need_qcspec_join:
            stmt = stmt.join(ManybodyRecordORM.specification)

        if need_qcspec_join:
            stmt = stmt.join(ManybodySpecificationORM.levels).join(
                ManybodySpecificationLevelsORM.singlepoint_specification
            )

        stmt = stmt.where(*and_query)

        return self.root_socket.records.query_base(
            stmt=stmt,
            orm_type=ManybodyRecordORM,
            query_data=query_data,
            session=session,
        )

    def add_internal(
        self,
        initial_molecule_ids: Sequence[int],
        mb_spec_id: int,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user_id: Optional[int],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Internal function for adding new manybody computations

        This function expects that the molecules and specification are already added to the
        database and that the ids are known.

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        initial_molecule_ids
            IDs of the initial molecules
        mb_spec_id
            ID of the specification
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user_id
            ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        compute_tag = compute_tag.lower()

        with self.root_socket.optional_session(session, False) as session:
            all_orm = []

            for mid in initial_molecule_ids:
                mb_orm = ManybodyRecordORM(
                    is_service=True,
                    specification_id=mb_spec_id,
                    initial_molecule_id=mid,
                    status=RecordStatusEnum.waiting,
                    creator_user_id=creator_user_id,
                )

                self.create_service(mb_orm, compute_tag, compute_priority, find_existing)
                all_orm.append(mb_orm)

            if find_existing:
                meta, ids = insert_general(
                    session,
                    all_orm,
                    (ManybodyRecordORM.specification_id, ManybodyRecordORM.initial_molecule_id),
                    (ManybodyRecordORM.id,),
                    use_unique=False,
                    lock_id=manybody_insert_lock_id,
                )
                return meta, [x[0] for x in ids]
            else:
                session.add_all(all_orm)
                session.flush()
                meta = InsertMetadata(inserted_idx=list(range(len(all_orm))))

                return meta, [x.id for x in all_orm]

    def add(
        self,
        initial_molecules: Sequence[Union[int, Molecule]],
        mb_spec: ManybodySpecification,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertMetadata, List[Optional[int]]]:
        """
        Adds new manybody calculations

        This checks if the calculations already exist in the database. If so, it returns
        the existing id, otherwise it will insert it and return the new id.

        Parameters
        ----------
        initial_molecules
            Initial molecules (objects or ids) to compute using the specification
        mb_spec
            Specification for the calculations
        compute_tag
            The tag for the task. This will assist in routing to appropriate compute managers.
        compute_priority
            The priority for the computation
        creator_user
            Name or ID of the user who created the record
        find_existing
            If True, search for existing records and return those. If False, always add new records
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about the insertion, and a list of record ids. The ids will be in the
            order of the input molecules
        """

        with self.root_socket.optional_session(session, False) as session:
            creator_user_id = self.root_socket.users.get_optional_user_id(creator_user, session=session)

            # First, add the specification
            spec_meta, spec_id = self.add_specification(mb_spec, session=session)
            if not spec_meta.success:
                return (
                    InsertMetadata(
                        error_description="Aborted - could not add specification: " + spec_meta.error_string
                    ),
                    [],
                )

            # Now the molecules
            mol_meta, mol_ids = self.root_socket.molecules.add_mixed(initial_molecules, session=session)
            if not mol_meta.success:
                return (
                    InsertMetadata(error_description="Aborted - could not add all molecules: " + mol_meta.error_string),
                    [],
                )

            return self.add_internal(
                mol_ids,
                spec_id,
                compute_tag,
                compute_priority,
                creator_user_id,
                find_existing,
                session=session,
            )

    def add_from_input(
        self,
        record_input: ManybodyInput,
        compute_tag: str,
        compute_priority: PriorityEnum,
        creator_user: Optional[Union[int, str]],
        find_existing: bool,
        *,
        session: Optional[Session] = None,
    ) -> Tuple[InsertCountsMetadata, int]:

        assert isinstance(record_input, ManybodyInput)

        meta, ids = self.add(
            [record_input.initial_molecule],
            record_input.specification,
            compute_tag,
            compute_priority,
            creator_user,
            find_existing,
        )

        return InsertCountsMetadata.from_insert_metadata(meta), ids[0]

    ####################################################
    # Some stuff to be retrieved for manybodys
    ####################################################

    def get_clusters(
        self,
        record_id: int,
        *,
        session: Optional[Session] = None,
    ) -> List[Dict[str, Any]]:
        options = [
            lazyload("*"),
            defer("*"),
            joinedload(ManybodyRecordORM.clusters).options(
                undefer("*"), joinedload(ManybodyClusterORM.molecule).options(undefer("*"))
            ),
        ]

        with self.root_socket.optional_session(session) as session:
            rec = session.get(ManybodyRecordORM, record_id, options=options)
            if rec is None:
                raise MissingDataError(f"Cannot find record {record_id}")
            return [x.model_dict() for x in rec.clusters]



from flask import current_app

from qcfractal.flask_app import storage_socket
from qcfractal.flask_app.api_v1.blueprint import api_v1
from qcfractal.flask_app.wrap_route import wrap_global_route
from qcportal.exceptions import InvalidUpdateError
from qcportal.internal_jobs import InternalJobQueryFilters, InternalJobStatusEnum
from qcportal.utils import calculate_limit


@api_v1.route("/internal_jobs/<int:job_id>", methods=["GET"])
@wrap_global_route("internal_jobs", "read")
def get_internal_job_v1(job_id: int):
    return storage_socket.internal_jobs.get(job_id=job_id)


@api_v1.route("/internal_jobs/<int:job_id>", methods=["DELETE"])
@wrap_global_route("internal_jobs", "delete")
def delete_internal_job_v1(job_id: int):
    return storage_socket.internal_jobs.delete(job_id=job_id)


@api_v1.route("/internal_jobs/<int:job_id>/status", methods=["PUT"])
@wrap_global_route("internal_jobs", "modify")
def cancel_internal_job_v1(job_id: int, body_data: InternalJobStatusEnum):
    if body_data == InternalJobStatusEnum.cancelled:
        return storage_socket.internal_jobs.cancel(job_id=job_id)
    else:
        raise InvalidUpdateError(f"Invalid status for updating internal job: {body_data}")


@api_v1.route("/internal_jobs/query", methods=["POST"])
@wrap_global_route("internal_jobs", "read")
def query_internal_jobs_v1(body_data: InternalJobQueryFilters):
    max_limit = current_app.config["QCFRACTAL_CONFIG"].api_limits.get_internal_jobs
    body_data.limit = calculate_limit(max_limit, body_data.limit)
    return storage_socket.internal_jobs.query(body_data)



"""remove role table

Revision ID: db730926b1b9
Revises: e15456ce68e2
Create Date: 2025-04-08 12:04:09.860028

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "db730926b1b9"
down_revision = "e15456ce68e2"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("user", sa.Column("role", sa.String(), nullable=True))
    op.execute(
        sa.text(
            """UPDATE "user" SET role = ri.rolename
                          FROM role ri
                          WHERE ri.id = "user".role_id"""
        )
    )

    op.alter_column("user", "role", nullable=False)

    op.drop_constraint("user_role_id_fkey", "user", type_="foreignkey")
    op.drop_table("role")
    op.drop_column("user", "role_id")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    raise NotImplementedError("Cannot downgrade")
    # ### end Alembic commands ###



"""add user preferences table

Revision ID: b664ee54c62a
Revises: 1136ef3ba3ef
Create Date: 2025-04-05 10:10:51.406526

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "b664ee54c62a"
down_revision = "1136ef3ba3ef"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "user_preferences",
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("preferences", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.ForeignKeyConstraint(["user_id"], ["user.id"], ondelete="cascade"),
        sa.PrimaryKeyConstraint("user_id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("user_preferences")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import TYPE_CHECKING

from sqlalchemy import (
    Column,
    Integer,
    TIMESTAMP,
    Float,
    Index,
    String,
    Enum,
    UniqueConstraint,
    CheckConstraint,
)
from sqlalchemy.dialects.postgresql import ARRAY, JSONB

from qcfractal.db_socket import BaseORM
from qcportal.managers import ManagerStatusEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from typing import Dict, Any, Optional, Iterable


class ComputeManagerORM(BaseORM):
    """
    Table for storing information about active and inactive compute managers
    """

    __tablename__ = "compute_manager"

    id = Column(Integer, primary_key=True)

    name = Column(String, nullable=False)
    cluster = Column(String, nullable=False)
    hostname = Column(String, nullable=False)
    username = Column(String)  # Can be null (ie, security not enabled, snowflake)
    compute_tags = Column(ARRAY(String), nullable=False)

    # Latest count
    claimed = Column(Integer, nullable=False, default=0)
    successes = Column(Integer, nullable=False, default=0)
    failures = Column(Integer, nullable=False, default=0)
    rejected = Column(Integer, nullable=False, default=0)

    active_tasks = Column(Integer, nullable=False, default=0)
    active_cores = Column(Integer, nullable=False, default=0)
    active_memory = Column(Float, nullable=False, default=0.0)
    total_cpu_hours = Column(Float, nullable=False, default=0.0)

    status = Column(Enum(ManagerStatusEnum), nullable=False)

    created_on = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)
    modified_on = Column(TIMESTAMP(timezone=True), nullable=False, default=now_at_utc)

    manager_version = Column(String, nullable=False)
    programs = Column(JSONB, nullable=False)

    __table_args__ = (
        Index("ix_compute_manager_status", "status"),
        Index("ix_compute_manager_modified_on", "modified_on", postgresql_using="brin"),
        UniqueConstraint("name", name="ux_compute_manager_name"),
        CheckConstraint("programs::text = LOWER(programs::text)", name="ck_compute_manager_programs_lower"),
        CheckConstraint("compute_tags::text = LOWER(compute_tags::text)", name="ck_compute_manager_compute_tags_lower"),
    )

    def model_dict(self, exclude: Optional[Iterable[str]] = None) -> Dict[str, Any]:
        d = BaseORM.model_dict(self, exclude)

        # TODO - DEPRECATED - remove eventually
        if "compute_tags" in d:
            d["tags"] = d.pop("compute_tags")

        return d



from qcarchivetesting import load_hash_test_data
from qcfractal.components.testing_fixtures import spec_test_runner
from qcfractal.components.torsiondrive.record_db_models import TorsiondriveSpecificationORM
from qcfractal.db_socket import SQLAlchemySocket
from qcportal.optimization import OptimizationSpecification, OptimizationProtocols
from qcportal.singlepoint import QCSpecification, SinglepointDriver, SinglepointProtocols
from qcportal.torsiondrive import TorsiondriveSpecification, TorsiondriveKeywords

common_opt_spec = OptimizationSpecification(
    program="optprog2",
    keywords={"k": "value"},
    protocols=OptimizationProtocols(trajectory="none"),
    qc_specification=QCSpecification(
        program="prog2",
        driver=SinglepointDriver.hessian,
        method="hf",
        basis="def2-tzvp",
        keywords={"k": "value"},
        protocols=SinglepointProtocols(wavefunction="orbitals_and_eigenvalues"),
    ),
)


def test_torsiondrive_hash_canaries(storage_socket: SQLAlchemySocket):
    # Test data is hash : spec dict
    test_data = load_hash_test_data("torsiondrive_specification_tests")

    # Hashes are independent of opt spec
    # TODO - protocols are not part of model
    for t in test_data.values():
        del t["protocols"]
    spec_map = [
        (k, TorsiondriveSpecification(optimization_specification=common_opt_spec, **v)) for k, v in test_data.items()
    ]

    specs = [x[1] for x in spec_map]
    meta, ids = storage_socket.records.torsiondrive.add_specifications(specs)
    assert meta.success
    assert len(ids) == len(specs)
    assert meta.n_existing == 0

    with storage_socket.session_scope() as session:
        for spec_id, (spec_hash, _) in zip(ids, spec_map):
            spec_orm = session.get(TorsiondriveSpecificationORM, spec_id)
            assert spec_orm.specification_hash == spec_hash


def test_torsiondrive_socket_add_specification_same_1(spec_test_runner):
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    spec_test_runner("torsiondrive", spec1, spec1, True)


def test_torsiondrive_socket_add_specification_same_2(spec_test_runner):
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    # model handling defaults
    spec2 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )
    spec_test_runner("torsiondrive", spec1, spec2, True)


def test_torsiondrive_socket_add_specification_same_3(spec_test_runner):
    # some changes to the opt spec
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    # model handling defaults
    spec2 = TorsiondriveSpecification(
        program="torSIOndrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            energy_upper_limit=0.05,
        ),
        optimization_specification=OptimizationSpecification(
            program="optpROg2",
            keywords={"k": "value"},
            protocols=OptimizationProtocols(trajectory="none"),
            qc_specification=QCSpecification(
                program="PROG2",
                driver=SinglepointDriver.gradient,
                method="hf",
                basis="def2-tZVp",
                keywords={"k": "value"},
                protocols=SinglepointProtocols(wavefunction="orbitals_and_eigenvalues"),
            ),
        ),
    )
    spec_test_runner("torsiondrive", spec1, spec2, True)


def test_torsiondrive_socket_add_specification_diff_1(spec_test_runner):
    #  changing energy upper limit
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    td_kw = spec1.keywords.copy(update={"energy_upper_limit": 0.051})
    spec2 = spec1.copy(update={"keywords": td_kw})
    spec_test_runner("torsiondrive", spec1, spec2, False)


def test_torsiondrive_socket_add_specification_diff_2(spec_test_runner):
    #  ordering of dihedrals
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    td_kw = spec1.keywords.copy(update={"dihedrals": [(8, 11, 13, 15)]})
    spec2 = spec1.copy(update={"keywords": td_kw})
    spec_test_runner("torsiondrive", spec1, spec2, False)


def test_torsiondrive_socket_add_specification_diff_3(spec_test_runner):
    #  grid spacing
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    td_kw = spec1.keywords.copy(update={"grid_spacing": [30]})
    spec2 = spec1.copy(update={"keywords": td_kw})
    spec_test_runner("torsiondrive", spec1, spec2, False)


def test_torsiondrive_socket_add_specification_diff_4(spec_test_runner):
    #  energy decrease thresh
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=None,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    td_kw = spec1.keywords.copy(update={"energy_decrease_thresh": 0.051})
    spec2 = spec1.copy(update={"keywords": td_kw})
    spec_test_runner("torsiondrive", spec1, spec2, False)


def test_torsiondrive_socket_add_specification_diff_5(spec_test_runner):
    # energy_upper_limit
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=0.2,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    td_kw = spec1.keywords.copy(update={"energy_upper_limit": 0.051})
    spec2 = spec1.copy(update={"keywords": td_kw})
    spec_test_runner("torsiondrive", spec1, spec2, False)


def test_torsiondrive_socket_add_specification_diff_6(spec_test_runner):
    # optimization keywords
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=0.2,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    opt_spec = spec1.optimization_specification.copy(update={"keywords": {"a": 1.0}})
    spec2 = spec1.copy(update={"optimization_specification": opt_spec})
    spec_test_runner("torsiondrive", spec1, spec2, False)


def test_torsiondrive_socket_add_specification_diff_7(spec_test_runner):
    # basis set
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=0.2,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    qc_spec = spec1.optimization_specification.qc_specification.copy(update={"basis": "def2-qzvp"})
    opt_spec = spec1.optimization_specification.copy(update={"qc_specification": qc_spec})
    spec2 = spec1.copy(update={"optimization_specification": opt_spec})
    spec_test_runner("torsiondrive", spec1, spec2, False)


def test_torsiondrive_socket_add_specification_diff_8(spec_test_runner):
    # qc keywords
    spec1 = TorsiondriveSpecification(
        program="torsiondrive",
        keywords=TorsiondriveKeywords(
            dihedrals=[(8, 11, 15, 13)],
            grid_spacing=[15],
            dihedral_ranges=None,
            energy_decrease_thresh=0.2,
            energy_upper_limit=0.05,
        ),
        optimization_specification=common_opt_spec,
    )

    qc_spec = spec1.optimization_specification.qc_specification.copy(update={"keywords": {"z": 1.0 - 10}})
    opt_spec = spec1.optimization_specification.copy(update={"qc_specification": qc_spec})
    spec2 = spec1.copy(update={"optimization_specification": opt_spec})
    spec_test_runner("torsiondrive", spec1, spec2, False)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcfractal.components.optimization.testing_helpers import (
    run_procedure_data as run_opt_procedure_data,
    submit_procedure_data as submit_opt_procedure_data,
)
from qcfractal.components.record_db_models import BaseRecordORM
from qcfractal.components.singlepoint.testing_helpers import submit_procedure_data as submit_sp_procedure_data
from qcfractal.components.testing_helpers import populate_records_status
from qcportal.managers import ManagerName
from qcportal.record_models import RecordStatusEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake
    from qcfractal.db_socket import SQLAlchemySocket
    from sqlalchemy.orm.session import Session


def test_record_socket_reset_assigned_manager(storage_socket: SQLAlchemySocket, session: Session):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")
    mname2 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="9876-5432-1098-7654")

    manager_programs = {"qcengine": ["unknown"], "psi4": ["unknown"], "geometric": ["v3.0"]}
    storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs=manager_programs,
        compute_tags=["tag1"],
    )
    storage_socket.managers.activate(
        name_data=mname2,
        manager_version="v2.0",
        username="bill",
        programs=manager_programs,
        compute_tags=["tag2"],
    )

    id_1, result_data_1 = submit_sp_procedure_data(storage_socket, "sp_psi4_water_energy", "tag1")
    id_2, result_data_2 = submit_sp_procedure_data(storage_socket, "sp_psi4_water_gradient", "tag2")
    id_3, result_data_3 = submit_sp_procedure_data(storage_socket, "sp_psi4_water_hessian", "tag1")
    id_4, result_data_4 = submit_opt_procedure_data(storage_socket, "opt_psi4_benzene", "tag2")
    id_5, result_data_5 = submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_1", "tag1")
    id_6, result_data_6 = submit_sp_procedure_data(storage_socket, "sp_psi4_benzene_energy_2", "tag1")
    all_id = [id_1, id_2, id_3, id_4, id_5, id_6]

    tasks_1 = storage_socket.tasks.claim_tasks(mname1.fullname, manager_programs, ["tag1"])
    tasks_2 = storage_socket.tasks.claim_tasks(mname2.fullname, manager_programs, ["tag2"])

    rec = [session.get(BaseRecordORM, i) for i in all_id]
    assert all(r.status == RecordStatusEnum.running for r in rec)
    assert all(r.task.available is False for r in rec)

    assert len(tasks_1) == 4
    assert len(tasks_2) == 2

    time_0 = now_at_utc()
    ids = storage_socket.records.reset_assigned(manager_name=[mname1.fullname])
    time_1 = now_at_utc()
    assert set(ids) == {id_1, id_3, id_5, id_6}

    session.expire_all()
    rec = [session.get(BaseRecordORM, i) for i in all_id]
    assert rec[0].status == RecordStatusEnum.waiting
    assert rec[1].status == RecordStatusEnum.running
    assert rec[2].status == RecordStatusEnum.waiting
    assert rec[3].status == RecordStatusEnum.running
    assert rec[4].status == RecordStatusEnum.waiting
    assert rec[5].status == RecordStatusEnum.waiting

    assert rec[0].manager_name is None
    assert rec[1].manager_name == mname2.fullname
    assert rec[2].manager_name is None
    assert rec[3].manager_name == mname2.fullname
    assert rec[4].manager_name is None
    assert rec[5].manager_name is None

    assert time_0 < rec[0].modified_on < time_1
    assert rec[1].modified_on < time_0
    assert time_0 < rec[2].modified_on < time_1
    assert rec[3].modified_on < time_0
    assert time_0 < rec[4].modified_on < time_1
    assert time_0 < rec[5].modified_on < time_1

    assert rec[0].task.available is True
    assert rec[1].task.available is False
    assert rec[2].task.available is True
    assert rec[3].task.available is False
    assert rec[4].task.available is True
    assert rec[5].task.available is True


def test_record_socket_reset_assigned_manager_none(storage_socket: SQLAlchemySocket):
    populate_records_status(storage_socket)
    ids = storage_socket.records.reset_assigned(manager_name=[])
    assert ids == []


def test_record_client_reset(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    all_id = populate_records_status(storage_socket)

    # Can reset only error
    time_0 = now_at_utc()
    meta = snowflake_client.reset_records(all_id)
    time_1 = now_at_utc()
    assert meta.n_updated == 1

    with storage_socket.session_scope() as session:
        rec = [session.get(BaseRecordORM, i) for i in all_id]

        # created_on shouldn't change
        for r in rec:
            assert r.created_on < time_0

        assert rec[0].status == RecordStatusEnum.waiting
        assert rec[1].status == RecordStatusEnum.complete
        assert rec[2].status == RecordStatusEnum.running
        assert rec[3].status == RecordStatusEnum.waiting
        assert rec[4].status == RecordStatusEnum.cancelled
        assert rec[5].status == RecordStatusEnum.deleted
        assert rec[6].status == RecordStatusEnum.invalid

        assert rec[0].task is not None
        assert rec[1].task is None
        assert rec[2].task is not None
        assert rec[3].task is not None
        assert rec[4].task is None
        assert rec[5].task is None
        assert rec[6].task is None

        assert rec[0].manager_name is None
        assert rec[1].manager_name is not None
        assert rec[2].manager_name is not None
        assert rec[3].manager_name is None
        assert rec[4].manager_name is None
        assert rec[5].manager_name is None
        assert rec[6].manager_name is not None

        assert rec[0].modified_on < time_0
        assert rec[1].modified_on < time_0
        assert rec[2].modified_on < time_0
        assert time_0 < rec[3].modified_on < time_1
        assert rec[4].modified_on < time_0
        assert rec[5].modified_on < time_0
        assert rec[6].modified_on < time_0


def test_record_socket_reset_none(storage_socket: SQLAlchemySocket):
    # The client may short-circuit with empty lists, so we should also test the socket separately
    populate_records_status(storage_socket)
    meta = storage_socket.records.reset([])
    assert meta.n_updated == 0


def test_record_client_reset_none(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.reset_records([])
    assert meta.n_updated == 0


def test_record_client_reset_missing(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    all_id = populate_records_status(storage_socket)
    meta = snowflake_client.reset_records([all_id[3], 9999])
    assert meta.success is False
    assert meta.n_updated == 1


def test_record_client_cancel(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    all_id = populate_records_status(storage_socket)

    # waiting, running, error can be cancelled
    time_0 = now_at_utc()
    meta = snowflake_client.cancel_records(all_id)
    time_1 = now_at_utc()
    assert meta.n_updated == 3

    with storage_socket.session_scope() as session:
        rec = [session.get(BaseRecordORM, i) for i in all_id]

        # created_on hasn't changed
        for r in rec:
            assert r.created_on < time_0

        assert rec[0].status == RecordStatusEnum.cancelled
        assert rec[1].status == RecordStatusEnum.complete
        assert rec[2].status == RecordStatusEnum.cancelled
        assert rec[3].status == RecordStatusEnum.cancelled
        assert rec[4].status == RecordStatusEnum.cancelled
        assert rec[5].status == RecordStatusEnum.deleted
        assert rec[6].status == RecordStatusEnum.invalid

        assert rec[0].task is None
        assert rec[1].task is None
        assert rec[2].task is None
        assert rec[3].task is None
        assert rec[4].task is None
        assert rec[5].task is None
        assert rec[6].task is None

        assert rec[0].manager_name is None
        assert rec[1].manager_name is not None
        assert rec[2].manager_name is None
        assert rec[3].manager_name is not None  # manager left for errored
        assert rec[4].manager_name is None
        assert rec[5].manager_name is None
        assert rec[6].manager_name is not None

        assert time_0 < rec[0].modified_on < time_1
        assert rec[1].modified_on < time_0
        assert time_0 < rec[2].modified_on < time_1
        assert time_0 < rec[3].modified_on < time_1
        assert rec[4].modified_on < time_0
        assert rec[5].modified_on < time_0
        assert rec[6].modified_on < time_0


def test_record_socket_cancel_none(storage_socket: SQLAlchemySocket):
    # The client may short-circuit with empty lists, so we should also test the socket separately
    populate_records_status(storage_socket)
    meta = storage_socket.records.cancel([])
    assert meta.n_updated == 0


def test_record_client_cancel_none(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.cancel_records([])
    assert meta.n_updated == 0


def test_record_client_cancel_missing(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    all_id = populate_records_status(storage_socket)
    meta = snowflake_client.cancel_records([all_id[0], 9999])
    assert meta.success is False
    assert meta.n_updated == 1


def test_record_client_invalidate(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    all_id = populate_records_status(storage_socket)

    # only completed can be invalidated
    time_0 = now_at_utc()
    meta = snowflake_client.invalidate_records(all_id)
    time_1 = now_at_utc()
    assert meta.n_updated == 1

    with storage_socket.session_scope() as session:
        rec = [session.get(BaseRecordORM, i) for i in all_id]

        # created_on hasn't changed
        for r in rec:
            assert r.created_on < time_0

        assert rec[0].status == RecordStatusEnum.waiting
        assert rec[1].status == RecordStatusEnum.invalid
        assert rec[2].status == RecordStatusEnum.running
        assert rec[3].status == RecordStatusEnum.error
        assert rec[4].status == RecordStatusEnum.cancelled
        assert rec[5].status == RecordStatusEnum.deleted
        assert rec[6].status == RecordStatusEnum.invalid

        assert rec[0].task is not None
        assert rec[1].task is None
        assert rec[2].task is not None
        assert rec[3].task is not None
        assert rec[4].task is None
        assert rec[5].task is None
        assert rec[6].task is None

        assert rec[0].task.available is True
        assert rec[2].task.available is False
        assert rec[3].task.available is False

        assert rec[0].manager_name is None
        assert rec[1].manager_name is not None  # Manager left on
        assert rec[2].manager_name is not None
        assert rec[3].manager_name is not None
        assert rec[4].manager_name is None
        assert rec[5].manager_name is None
        assert rec[6].manager_name is not None

        assert rec[0].modified_on < time_0
        assert time_0 < rec[1].modified_on < time_1
        assert rec[2].modified_on < time_0
        assert rec[3].modified_on < time_0
        assert rec[4].modified_on < time_0
        assert rec[5].modified_on < time_0
        assert rec[6].modified_on < time_0


def test_record_socket_invalidate_none(storage_socket: SQLAlchemySocket):
    # The client may short-circuit with empty lists, so we should also test the socket separately
    populate_records_status(storage_socket)
    meta = storage_socket.records.invalidate([])
    assert meta.n_updated == 0


def test_record_client_invalidate_none(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.invalidate_records([])
    assert meta.n_updated == 0


def test_record_client_invalidate_missing(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    all_id = populate_records_status(storage_socket)
    meta = snowflake_client.invalidate_records([all_id[1], 9999])
    assert meta.success is False
    assert meta.n_updated == 1


def test_record_client_softdelete(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    all_id = populate_records_status(storage_socket)

    # only deleted can't be deleted
    time_0 = now_at_utc()
    meta = snowflake_client.delete_records(all_id, soft_delete=True)
    time_1 = now_at_utc()
    assert meta.n_deleted == 6
    assert meta.deleted_idx == [0, 1, 2, 3, 4, 6]
    assert meta.error_idx == [5]  # deleted can't be deleted

    with storage_socket.session_scope() as session:
        rec = [session.get(BaseRecordORM, i) for i in all_id]

        # created_on hasn't changed
        for r in rec:
            assert r.created_on < time_0

            assert r.status == RecordStatusEnum.deleted
            assert r.task is None

        assert time_0 < rec[0].modified_on < time_1
        assert time_0 < rec[1].modified_on < time_1
        assert time_0 < rec[2].modified_on < time_1
        assert time_0 < rec[3].modified_on < time_1
        assert time_0 < rec[4].modified_on < time_1
        assert rec[5].modified_on < time_0
        assert time_0 < rec[6].modified_on < time_1

        # completed and errored records should keep their manager
        assert rec[0].manager_name is None
        assert rec[1].manager_name is not None
        assert rec[2].manager_name is None
        assert rec[3].manager_name is not None
        assert rec[4].manager_name is None
        assert rec[5].manager_name is None
        assert rec[6].manager_name is not None


def test_record_client_softdelete_missing(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    all_id = populate_records_status(storage_socket)
    meta = snowflake_client.delete_records(all_id + [99999], soft_delete=True)
    assert meta.success is False
    assert meta.deleted_idx == [0, 1, 2, 3, 4, 6]
    assert meta.n_deleted == 6
    assert meta.error_idx == [5, 7]


def test_record_socket_softdelete_none(storage_socket: SQLAlchemySocket):
    # The client may short-circuit with empty lists, so we should also test the socket separately
    populate_records_status(storage_socket)
    meta = storage_socket.records.delete([], soft_delete=True)
    assert meta.success is True
    assert meta.n_deleted == 0


def test_record_client_softdelete_none(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.delete_records([], soft_delete=True)
    assert meta.success is True
    assert meta.n_deleted == 0


def test_record_client_harddelete_1(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    all_id = populate_records_status(storage_socket)

    # only deleted can't be deleted
    meta = snowflake_client.delete_records(all_id, soft_delete=False)
    assert meta.success
    assert meta.deleted_idx == [0, 1, 2, 3, 4, 5, 6]
    assert meta.n_deleted == 7

    with storage_socket.session_scope() as session:
        rec = [session.get(BaseRecordORM, i) for i in all_id]
        assert all(x is None for x in rec)


def test_record_client_harddelete_2(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    # Delete only some records
    all_id = populate_records_status(storage_socket)

    # only deleted can't be deleted
    meta = snowflake_client.delete_records([all_id[0], all_id[4]], soft_delete=False)
    assert meta.success
    assert meta.deleted_idx == [0, 1]
    assert meta.n_deleted == 2

    with storage_socket.session_scope() as session:
        rec = [session.get(BaseRecordORM, i) for i in all_id]
        assert rec[0] is None
        assert rec[1] is not None
        assert rec[2] is not None
        assert rec[3] is not None
        assert rec[4] is None
        assert rec[5] is not None
        assert rec[6] is not None


def test_record_socket_harddelete_none(storage_socket: SQLAlchemySocket):
    # The client may short-circuit with empty lists, so we should also test the socket separately
    populate_records_status(storage_socket)
    meta = storage_socket.records.delete([], soft_delete=False)
    assert meta.success is True
    assert meta.deleted_idx == []


def test_record_client_harddelete_none(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.delete_records([], soft_delete=False)
    assert meta.success is True
    assert meta.deleted_idx == []


def test_record_client_harddelete_missing(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    # Deletion always succeeds even if record doesn't exist
    all_id = populate_records_status(storage_socket)
    meta = snowflake_client.delete_records(all_id + [99999], soft_delete=False)
    assert meta.success is True
    assert meta.deleted_idx == [0, 1, 2, 3, 4, 5, 6, 7]
    assert meta.n_deleted == 8


def test_record_client_revert_chain(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    # Tests undelete, uninvalidate, uncancel
    all_id = populate_records_status(storage_socket)

    # cancel, invalidate, then delete all
    meta = snowflake_client.cancel_records(all_id)
    assert meta.n_updated == 3

    meta = snowflake_client.invalidate_records(all_id)
    assert meta.n_updated == 1

    meta = snowflake_client.delete_records(all_id)
    assert meta.n_deleted == 6

    with storage_socket.session_scope() as session:
        rec = [session.get(BaseRecordORM, i) for i in all_id]
        assert len(rec[0].info_backup) == 2
        assert len(rec[1].info_backup) == 2
        assert len(rec[2].info_backup) == 2
        assert len(rec[3].info_backup) == 2
        assert len(rec[4].info_backup) == 2
        assert len(rec[5].info_backup) == 1  # deleted in populate_db
        assert len(rec[6].info_backup) == 2

        meta = snowflake_client.undelete_records(all_id)
        assert meta.n_updated == 7

        session.expire_all()
        rec = [session.get(BaseRecordORM, i) for i in all_id]
        assert rec[0].status == RecordStatusEnum.cancelled
        assert rec[1].status == RecordStatusEnum.invalid
        assert rec[2].status == RecordStatusEnum.cancelled
        assert rec[3].status == RecordStatusEnum.cancelled
        assert rec[4].status == RecordStatusEnum.cancelled
        assert rec[5].status == RecordStatusEnum.waiting  # from populate_db
        assert rec[6].status == RecordStatusEnum.invalid

        assert rec[0].task is None
        assert rec[1].task is None
        assert rec[2].task is None
        assert rec[3].task is None
        assert rec[4].task is None
        assert rec[5].task is not None
        assert rec[6].task is None

        assert rec[5].task.available is True

        assert len(rec[0].info_backup) == 1
        assert len(rec[1].info_backup) == 1
        assert len(rec[2].info_backup) == 1
        assert len(rec[3].info_backup) == 1
        assert len(rec[4].info_backup) == 1
        assert len(rec[5].info_backup) == 0
        assert len(rec[6].info_backup) == 1

        meta = snowflake_client.uncancel_records(all_id)
        assert meta.n_updated == 4

        session.expire_all()
        rec = [session.get(BaseRecordORM, i) for i in all_id]
        assert rec[0].status == RecordStatusEnum.waiting
        assert rec[1].status == RecordStatusEnum.invalid
        assert rec[2].status == RecordStatusEnum.waiting
        assert rec[3].status == RecordStatusEnum.error
        assert rec[4].status == RecordStatusEnum.waiting  # from populate_db
        assert rec[5].status == RecordStatusEnum.waiting  # from populate_db
        assert rec[6].status == RecordStatusEnum.invalid

        assert len(rec[0].info_backup) == 0
        assert len(rec[1].info_backup) == 1
        assert len(rec[2].info_backup) == 0
        assert len(rec[3].info_backup) == 0
        assert len(rec[4].info_backup) == 0
        assert len(rec[5].info_backup) == 0
        assert len(rec[6].info_backup) == 1

        assert rec[0].task is not None
        assert rec[1].task is None
        assert rec[2].task is not None
        assert rec[3].task is not None
        assert rec[4].task is not None
        assert rec[5].task is not None
        assert rec[6].task is None

        assert rec[0].task.available is True
        assert rec[2].task.available is True
        assert rec[3].task.available is False
        assert rec[4].task.available is True
        assert rec[5].task.available is True

        meta = snowflake_client.uninvalidate_records(all_id)
        assert meta.n_updated == 2

        session.expire_all()
        rec = [session.get(BaseRecordORM, i) for i in all_id]
        assert rec[0].status == RecordStatusEnum.waiting
        assert rec[1].status == RecordStatusEnum.complete
        assert rec[2].status == RecordStatusEnum.waiting
        assert rec[3].status == RecordStatusEnum.error
        assert rec[4].status == RecordStatusEnum.waiting  # from populate_db
        assert rec[5].status == RecordStatusEnum.waiting  # from populate_db
        assert rec[6].status == RecordStatusEnum.complete

        assert len(rec[0].info_backup) == 0
        assert len(rec[1].info_backup) == 0
        assert len(rec[2].info_backup) == 0
        assert len(rec[3].info_backup) == 0
        assert len(rec[4].info_backup) == 0
        assert len(rec[5].info_backup) == 0
        assert len(rec[6].info_backup) == 0

        assert rec[0].task is not None
        assert rec[1].task is None
        assert rec[2].task is not None
        assert rec[3].task is not None
        assert rec[4].task is not None
        assert rec[5].task is not None
        assert rec[6].task is None

        assert rec[0].task.available is True
        assert rec[2].task.available is True
        assert rec[3].task.available is False
        assert rec[4].task.available is True
        assert rec[5].task.available is True


def test_record_socket_undelete_none(storage_socket: SQLAlchemySocket):
    populate_records_status(storage_socket)
    meta = storage_socket.records.undelete([])
    assert meta.success is True
    assert meta.updated_idx == []
    assert meta.n_updated == 0


def test_record_client_undelete_none(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.undelete_records([])
    assert meta.success is True
    assert meta.updated_idx == []
    assert meta.n_updated == 0


def test_record_client_undelete_missing(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.undelete_records([99999])
    assert meta.success is False
    assert meta.updated_idx == []
    assert meta.n_updated == 0
    assert meta.error_idx == [0]


def test_record_socket_uncancel_none(storage_socket: SQLAlchemySocket):
    populate_records_status(storage_socket)
    meta = storage_socket.records.uncancel([])
    assert meta.success is True
    assert meta.updated_idx == []
    assert meta.n_updated == 0


def test_record_client_uncancel_none(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.uncancel_records([])
    assert meta.success is True
    assert meta.updated_idx == []
    assert meta.n_updated == 0


def test_record_client_uncancel_missing(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.uncancel_records([99999])
    assert meta.success is False
    assert meta.updated_idx == []
    assert meta.n_updated == 0
    assert meta.error_idx == [0]


def test_record_socket_uninvalidate_none(storage_socket: SQLAlchemySocket):
    populate_records_status(storage_socket)
    meta = storage_socket.records.uninvalidate([])
    assert meta.success is True
    assert meta.updated_idx == []
    assert meta.n_updated == 0


def test_record_client_uninvalidate_none(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.uninvalidate_records([])
    assert meta.success is True
    assert meta.updated_idx == []
    assert meta.n_updated == 0


def test_record_client_uninvalidate_missing(snowflake: QCATestingSnowflake):
    storage_socket = snowflake.get_storage_socket()
    snowflake_client = snowflake.client()

    populate_records_status(storage_socket)
    meta = snowflake_client.uninvalidate_records([99999])
    assert meta.success is False
    assert meta.updated_idx == []
    assert meta.n_updated == 0
    assert meta.error_idx == [0]


@pytest.mark.parametrize("opt_file", ["opt_psi4_benzene", "opt_psi4_fluoroethane_notraj"])
def test_record_client_delete_children(snowflake: QCATestingSnowflake, opt_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    # Deleting with deleting children
    id1 = run_opt_procedure_data(storage_socket, activated_manager_name, opt_file)

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, id1)
        child_ids = [x.singlepoint_id for x in rec.trajectory]

        meta = snowflake_client.delete_records([id1], soft_delete=True, delete_children=True)
        assert meta.success
        assert meta.deleted_idx == [0]
        assert meta.n_children_deleted == len(child_ids)

        session.expire_all()
        child_recs = [session.get(BaseRecordORM, i) for i in child_ids]
        assert all(x.status == RecordStatusEnum.deleted for x in child_recs)

        meta = snowflake_client.delete_records([id1], soft_delete=False, delete_children=True)
        assert meta.success
        assert meta.deleted_idx == [0]
        assert meta.n_children_deleted == len(child_ids)

        session.expire_all()
        rec = session.get(BaseRecordORM, id1)
        assert rec is None

        child_recs = [session.get(BaseRecordORM, i) for i in child_ids]
        assert all(x is None for x in child_recs)


@pytest.mark.parametrize("opt_file", ["opt_psi4_benzene", "opt_psi4_fluoroethane_notraj"])
def test_record_client_delete_nochildren(snowflake: QCATestingSnowflake, opt_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    # Deleting without deleting children
    id1 = run_opt_procedure_data(storage_socket, activated_manager_name, opt_file)

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, id1)
        child_ids = [x.singlepoint_id for x in rec.trajectory]

        meta = snowflake_client.delete_records([id1], soft_delete=True, delete_children=False)
        assert meta.success
        assert meta.deleted_idx == [0]
        assert meta.n_children_deleted == 0

        session.expire_all()
        child_recs = [session.get(BaseRecordORM, i) for i in child_ids]
        assert all(x.status == RecordStatusEnum.complete for x in child_recs)

        meta = snowflake_client.delete_records([id1], soft_delete=False, delete_children=False)
        assert meta.success
        assert meta.deleted_idx == [0]
        assert meta.n_children_deleted == 0

        session.expire_all()
        rec = session.get(BaseRecordORM, id1)
        assert rec is None

        child_recs = [session.get(BaseRecordORM, i) for i in child_ids]
        assert all(x.status == RecordStatusEnum.complete for x in child_recs)


@pytest.mark.parametrize("opt_file", ["opt_psi4_benzene", "opt_psi4_fluoroethane_notraj"])
def test_record_client_undelete_children(snowflake: QCATestingSnowflake, opt_file: str):
    storage_socket = snowflake.get_storage_socket()
    activated_manager_name, _ = snowflake.activate_manager()
    snowflake_client = snowflake.client()

    # Deleting with deleting children, then undeleting
    id1 = run_opt_procedure_data(storage_socket, activated_manager_name, opt_file)

    with storage_socket.session_scope() as session:
        rec = session.get(BaseRecordORM, id1)
        child_ids = [x.singlepoint_id for x in rec.trajectory]

        meta = snowflake_client.delete_records([id1], soft_delete=True, delete_children=True)
        assert meta.success
        assert meta.deleted_idx == [0]
        assert meta.n_children_deleted == len(child_ids)

        meta = snowflake_client.undelete_records([id1])
        assert meta.success
        assert meta.updated_idx == [0]

        child_recs = [session.get(BaseRecordORM, i) for i in child_ids]
        assert all(x.status == RecordStatusEnum.complete for x in child_recs)



"""
The global qcfractal config file specification.
"""

from __future__ import annotations

import logging
import re
import os
import secrets
import tempfile
from typing import Optional, Dict, Union, List, Any

import yaml
from psycopg2.extensions import make_dsn, parse_dsn
from pydantic import BaseModel, Field, field_validator, model_validator, ValidationError, ConfigDict
from pydantic_settings import BaseSettings, SettingsConfigDict, PydanticBaseSettingsSource
from sqlalchemy.engine.url import URL, make_url

from qcfractal.port_util import find_open_port
from qcportal.utils import duration_to_seconds, update_nested_dict


def _make_abs_path(path: Optional[str], base_folder: str, default_filename: Optional[str]) -> Optional[str]:
    # No path specified, no default
    if path is None and default_filename is None:
        return None

    # Path isn't specified, but default is given
    if path is None:
        path = default_filename

    path = os.path.expanduser(path)
    if os.path.isabs(path):
        return path
    else:
        path = os.path.join(base_folder, path)
        return os.path.abspath(path)


def make_uri_string(
    host: Optional[str],
    port: Optional[Union[int, str]],
    username: Optional[str],
    password: Optional[str],
    dbname: Optional[str],
    query: Optional[Dict[str, str]],
) -> str:
    username = username if username is not None else ""
    password = ":" + password if password is not None else ""
    sep = "@" if username != "" or password != "" else ""
    query_str = "" if query is None else "&".join(f"{k}={v}" for k, v in query.items())

    # If this is a socket file, move the host to the query params
    if host.startswith("/"):
        query_str = "&" + query_str if query_str != "" else ""
        return f"postgresql://{username}{password}{sep}:{port}/{dbname}?host={host}{query_str}"
    else:
        query_str = "?" + query_str if query_str != "" else ""
        return f"postgresql://{username}{password}{sep}{host}:{port}/{dbname}{query_str}"


class QCFConfigBase(BaseModel):
    model_config = ConfigDict(extra="forbid")


class DatabaseConfig(QCFConfigBase):
    """
    Settings for the database used by QCFractal
    """

    base_folder: str = Field(
        ...,
        description="The base folder to use as the default for some options (logs, etc). Default is the location of the config file.",
    )

    full_uri: Optional[str] = Field(
        None, description="Full connection URI. This overrides host,username,password,port, etc"
    )

    host: str = Field(
        "localhost",
        description="The hostname or ip address the database is running on. If own = True, this must be localhost. May also be a path to a directory containing the database socket file",
    )
    port: int = Field(
        5432,
        description="The port the database is running on. If own = True, a database will be started, binding to this port",
    )
    database_name: str = Field("qcfractal_default", description="The database name to connect to.")
    username: str = Field(..., description="The database username to connect with")
    password: str = Field(..., description="The database password to connect with")
    query: Dict[str, Union[str, int]] = Field(
        {}, description="Extra connection query parameters at the end of the URL string"
    )

    own: bool = Field(
        True,
        description="If True, QCFractal will control the database instance. If False, you must start and manage the database yourself",
    )

    data_directory: Optional[str] = Field(
        None,
        description="Location to place the database if own == True. Default is [base_folder]/database if we own the database",
    )
    logfile: Optional[str] = Field(
        None,
        description="Path to a file to use as the database logfile (if own == True). Default is [base_folder]/qcfractal_database.log",
    )

    echo_sql: bool = Field(False, description="[ADVANCED] output raw SQL queries being run")
    pg_tool_dir: Optional[str] = Field(
        None,
        description="Directory containing Postgres tools such as psql and pg_ctl (ie, /usr/bin, or /usr/lib/postgresql/13/bin). If not specified, an attempt to find them will be made. This field is only required if autodetection fails and own == True",
    )

    pool_size: int = Field(
        5,
        description="[ADVANCED] set the size of the connection pool to use in SQLAlchemy. Set to zero to disable pooling",
    )

    maintenance_db: str = Field(
        "postgres",
        description="[ADVANCED] An existing database (not the one you want to use/create). This is used for database management",
    )

    @model_validator(mode="after")
    def _check_paths(self):
        if self.own:
            self.data_directory = _make_abs_path(self.data_directory, self.base_folder, "postgres")
            self.logfile = _make_abs_path(self.logfile, self.base_folder, "qcfractal_database.log")
        return self

    @property
    def database_uri(self) -> str:
        """
        Returns the real database URI as a string

        It does not hide the password, so is not suitable for logging
        """
        if self.full_uri is not None:
            return self.full_uri
        else:
            return make_uri_string(
                host=self.host,
                port=self.port,
                username=self.username,
                password=self.password,
                dbname=self.database_name,
                query=self.query,
            )

    @property
    def sqlalchemy_url(self) -> URL:
        """Returns the SQLAlchemy URL for this database"""

        url = make_url(self.database_uri)
        return url.set(drivername="postgresql+psycopg2")

    @property
    def psycopg2_dsn(self) -> str:
        """
        Returns a string suitable for use as a psycopg2 connection string
        """
        dsn_dict = parse_dsn(self.database_uri)
        return make_dsn(**dsn_dict)

    @property
    def psycopg2_maintenance_dsn(self) -> str:
        dsn_dict = parse_dsn(self.database_uri)
        dsn_dict["dbname"] = self.maintenance_db
        return make_dsn(**dsn_dict)

    @property
    def safe_uri(self) -> str:
        """
        Returns a user-readable version of the URI for logging, etc.
        """

        dsn = parse_dsn(self.database_uri)

        host = dsn.pop("host")
        port = dsn.pop("port", None)
        user = dsn.pop("user", None)
        password = dsn.pop("password", None)
        dbname = dsn.pop("dbname")

        # SQLAlchemy render_string has some problems sometimes, so use our own
        return make_uri_string(
            host=host, port=port, username=user, password="********" if password else None, dbname=dbname, query=dsn
        )  # everything left over


class AutoResetConfig(QCFConfigBase):
    """
    Limits on the number of records returned per query. This can be specified per object (molecule, etc)
    """

    enabled: bool = Field(False, description="Enable/disable automatic restart. True = enabled")
    unknown_error: int = Field(2, description="Max restarts for unknown errors")
    compute_lost: int = Field(5, description="Max restarts for computations where the compute resource disappeared")
    random_error: int = Field(5, description="Max restarts for random errors")


class APILimitConfig(QCFConfigBase):
    """
    Limits on the number of records returned per query. This can be specified per object (molecule, etc)
    """

    get_records: int = Field(1000, description="Number of calculation records that can be retrieved")
    add_records: int = Field(500, description="Number of calculation records that can be added")

    get_dataset_entries: int = Field(2000, description="Number of dataset entries that can be retrieved")

    get_molecules: int = Field(1000, description="Number of molecules that can be retrieved")
    add_molecules: int = Field(1000, description="Number of molecules that can be added")

    get_managers: int = Field(1000, description="Number of manager records to return")

    manager_tasks_claim: int = Field(200, description="Number of tasks a single manager can pull down")
    manager_tasks_return: int = Field(10, description="Number of tasks a single manager can return at once")

    get_access_logs: int = Field(1000, description="Number of access log records to return")
    get_error_logs: int = Field(100, description="Number of error log records to return")
    get_internal_jobs: int = Field(1000, description="Number of internal jobs to return")


class WebAPIConfig(QCFConfigBase):
    """
    Settings for the Web API (api) interface
    """

    num_threads_per_worker: int = Field(4, description="Number of threads per worker")
    worker_timeout: int = Field(
        120,
        description="If the master process does not hear from a worker for the given amount of time (in seconds),"
        "kill it. This effectively limits the time a worker has to respond to a request",
    )
    host: str = Field("localhost", description="The IP address or hostname to bind to")
    port: int = Field(7777, description="The port on which to run the REST interface.")

    secret_key: str = Field(..., description="Secret key for flask api. See documentation")
    jwt_secret_key: str = Field(..., description="Secret key for web tokens. See documentation")
    jwt_access_token_expires: int = Field(
        60 * 60, description="The time (in seconds) an access token is valid for. Default is 1 hour"
    )
    jwt_refresh_token_expires: int = Field(
        60 * 60 * 24, description="The time (in seconds) a refresh token is valid for. Default is 1 day"
    )
    user_session_max_age: int = Field(
        60 * 60 * 24, description="The time (in seconds) that a user session can be idle (for browser-based sessions)"
    )
    user_session_cookie_name: str = Field(
        "qcf_session", description="Name to use for a session cookie (for browser-based sessions)"
    )
    user_session_cookie_domain: Optional[str] = Field(
        None, description="Domain to use for the user-session cookie (for browser-based sessions)"
    )
    user_session_cookie_samesite: Optional[str] = Field(
        None, description="Set the SameSite flag for the user-session cookie (for browser-based sessions)"
    )
    user_session_cookie_partitioned: bool = Field(
        False, description="Use the Partitioned flag for the user-session cookie (for browser-based sessions)"
    )
    user_session_cookie_secure: bool = Field(
        False, description="Use Secure flag for the user-session cookie (for browser-based sessions)"
    )
    user_session_cookie_httponly: bool = Field(
        False, description="Use Secure flag for the user-session cookie (for browser-based sessions)"
    )

    extra_flask_options: Optional[Dict[str, Any]] = Field(
        None, description="Any additional options to pass directly to flask"
    )
    extra_waitress_options: Optional[Dict[str, Any]] = Field(
        None, description="Any additional options to pass directly to the waitress serve function"
    )

    @field_validator(
        "jwt_access_token_expires",
        "jwt_refresh_token_expires",
        "user_session_max_age",
        mode="before",
    )
    @classmethod
    def _convert_durations(cls, v):
        return duration_to_seconds(v)


class S3BucketMap(QCFConfigBase):
    dataset_attachment: str = Field("dataset_attachment", description="Bucket to hold dataset views")


class S3Config(QCFConfigBase):
    """
    Settings for using external files with S3
    """

    enabled: bool = False
    verify: bool = True
    passthrough: bool = False
    endpoint_url: Optional[str] = Field(None, description="S3 endpoint URL")
    access_key_id: Optional[str] = Field(None, description="AWS/S3 access key")
    secret_access_key: Optional[str] = Field(None, description="AWS/S3 secret key")

    bucket_map: S3BucketMap = Field(
        default_factory=S3BucketMap, description="Configuration for where to store various files"
    )

    @model_validator(mode="after")
    def _check_enabled(self):
        if self.enabled:
            for key in ["endpoint_url", "access_key_id", "secret_access_key"]:
                if getattr(self, key) is None:
                    raise ValueError(f"S3 enabled but {key} not set")
        return self


class CORSconfig(QCFConfigBase):
    """
    Settings for using CORS
    """

    enabled: bool = False
    origins: List[str] = Field([])
    supports_credentials: bool = False
    headers: List[str] = Field([])
    methods: List[str] = Field([])


class FractalConfig(BaseSettings):
    """
    Fractal Server settings
    """

    base_folder: str = Field(
        ...,
        description="The base directory to use as the default for some options (logs, etc). Default is the location of the config file.",
    )

    temporary_dir: Optional[str] = Field(
        None,
        description="Temporary directory to use for things such as view creation. If None, uses system default. This may require a lot of space!",
    )

    # Info for the REST interface
    name: str = Field("QCFractal Server", description="The QCFractal server name")

    enable_security: bool = Field(True, description="Enable user authentication and authorization")
    allow_unauthenticated_read: bool = Field(
        True,
        description="Allows unauthenticated read access to this instance. This does not extend to sensitive tables (such as user information)",
    )
    strict_compute_tags: bool = Field(
        False,
        description="If True, disables wildcard behavior for compute tags. This disables managers from claiming all "
        "tags if they specify a wildcard ('*') tag. Managers will still be able to claim tasks with an "
        "explicit '*' tag if they specify the '*' queue tag in their config",
    )

    # Logging and profiling
    logfile: Optional[str] = Field(
        None,
        description="Path to a file to use for server logging. If not specified, logs will be printed to standard output",
    )
    loglevel: str = Field(
        "INFO", description="Level of logging to enable (debug, info, warning, error, critical). Case insensitive"
    )

    hide_internal_errors: bool = Field(
        True,
        description="If True, internal errors will only be reported as an error "
        "number to the user. If False, the entire error/backtrace "
        "will be sent (which could rarely contain sensitive info). "
        "In either case, errors will be stored in the database",
    )

    # Periodics
    service_frequency: int = Field(60, description="The frequency at which to update services (in seconds)")
    max_active_services: int = Field(20, description="The maximum number of concurrent active services")
    heartbeat_frequency: int = Field(
        1800,
        description="The frequency (in seconds) to check the heartbeat of compute managers",
        gt=0,
    )
    heartbeat_frequency_jitter: float = Field(
        0.1, description="Jitter fraction to be applied to the heartbeat frequency", ge=0
    )
    heartbeat_max_missed: int = Field(
        5,
        description="The maximum number of heartbeats that a compute manager can miss. If more are missed, the worker is considered dead",
        ge=0,
    )

    # Access logging
    log_access: bool = Field(False, description="Store API access in the database")
    access_log_keep: int = Field(
        0, description="How far back to keep access logs (in days or as a duration string). 0 means keep all"
    )

    # maxmind_account_id: Optional[int] = Field(None, description="Account ID for MaxMind GeoIP2 service")
    maxmind_license_key: Optional[str] = Field(
        None,
        description="License key for MaxMind GeoIP2 service. If provided, the GeoIP2 database will be downloaded and updated automatically",
    )

    geoip2_dir: Optional[str] = Field(
        None,
        description="Directory containing the Maxmind GeoIP2 Cities file (GeoLite2-City.mmdb) Defaults to [base_folder]/geoip2. This directory will be created if needed.",
    )

    geoip2_filename: str = Field(
        "GeoLite2-City.mmdb", description="Filename of the Maxmind GeoIP2 Cities file (GeoLite2-City.mmdb)"
    )

    # Internal jobs
    internal_job_processes: int = Field(
        1, description="Number of processes for processing internal jobs and async requests"
    )
    internal_job_keep: int = Field(
        0, description="How far back to keep finished internal jobs (in days or as a duration string). 0 means keep all"
    )

    # Homepage settings
    homepage_redirect_url: Optional[str] = Field(None, description="Redirect to this URL when going to the root path")
    homepage_directory: Optional[str] = Field(None, description="Use this directory to serve the homepage")

    # File uploads
    upload_directory: Optional[str] = Field(None, description="Directory to store user-uploaded files for processing")

    # Other settings blocks
    database: DatabaseConfig = Field(..., description="Configuration of the settings for the database")
    api: WebAPIConfig = Field(..., description="Configuration of the REST interface")
    s3: S3Config = Field(default_factory=S3Config, description="Configuration of the S3 file storage (optional)")
    api_limits: APILimitConfig = Field(
        default_factory=APILimitConfig, description="Configuration of the limits to the api"
    )
    cors: CORSconfig = Field(
        default_factory=CORSconfig, description="Configuration Cross Origin Resource sharing (advanced)"
    )
    auto_reset: AutoResetConfig = Field(
        default_factory=AutoResetConfig, description="Configuration for automatic resetting of tasks"
    )

    @field_validator("loglevel", mode="after")
    @classmethod
    def _check_loglevel(cls, v):
        v = v.upper()
        if v not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
            raise ValidationError(f"{v} is not a valid loglevel. Must be DEBUG, INFO, WARNING, ERROR, or CRITICAL")
        return v

    @field_validator("service_frequency", "heartbeat_frequency", mode="before")
    @classmethod
    def _convert_durations(cls, v):
        return duration_to_seconds(v)

    @field_validator("access_log_keep", "internal_job_keep", mode="before")
    def _convert_durations_days(cls, v):
        if isinstance(v, int) or (isinstance(v, str) and v.isdigit()):
            return int(v) * 86400
        return duration_to_seconds(v)

    @model_validator(mode="before")
    @classmethod
    def _propagate_base_folder(cls, values):
        if isinstance(values, dict) and "base_folder" in values:
            values.setdefault("database", {})
            values["database"]["base_folder"] = values["base_folder"]
        return values

    @model_validator(mode="before")
    @classmethod
    def _detect_deprecated_fields(cls, values):
        if isinstance(values, dict):
            if "strict_queue_tags" in values:
                logging.getLogger(__name__).warning("strict_queue_tags is deprecated. Use strict_compute_tags instead")
                values["strict_compute_tags"] = values.pop("strict_queue_tags")
        return values

    @model_validator(mode="after")
    def _check_paths(self):
        self.homepage_directory = _make_abs_path(self.homepage_directory, self.base_folder, None)
        self.upload_directory = _make_abs_path(self.upload_directory, self.base_folder, None)
        self.logfile = _make_abs_path(self.logfile, self.base_folder, None)
        self.geoip2_dir = _make_abs_path(self.geoip2_dir, self.base_folder, "geoip2")

        if self.temporary_dir is None:
            self.temporary_dir = tempfile.gettempdir()
            self.temporary_dir = _make_abs_path("qcf_tmp", self.base_folder, tempfile.gettempdir())
        else:
            self.temporary_dir = _make_abs_path(self.temporary_dir, self.base_folder, None)

        os.makedirs(self.temporary_dir, exist_ok=True)
        return self

    model_config = SettingsConfigDict(
        extra="forbid", case_sensitive=False, env_prefix="QCF_", env_nested_delimiter="__"
    )

    # Since we manually read the yaml files, the values passed into the init
    # come from files and should be lower priority that env settings
    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return env_settings, dotenv_settings, init_settings, file_secret_settings


def read_configuration(file_paths: list[str], extra_config: Optional[Dict[str, Any]] = None) -> FractalConfig:
    """
    Reads QCFractal configuration from YAML files
    """
    logger = logging.getLogger(__name__)
    config_data: Dict[str, Any] = {}

    # Read all the files, in order
    for path in file_paths:
        with open(path, "r") as yf:
            logger.info(f"Reading configuration data from {path}")
            file_data = yaml.safe_load(yf)
            update_nested_dict(config_data, file_data)

    if extra_config:
        update_nested_dict(config_data, extra_config)

    # Find the base folder
    # 1. If specified in the environment, use that
    # 2. Use any specified in a config file
    # 3. Use the path of the last (highest-priority) file given
    if "QCF_BASE_FOLDER" in os.environ:
        base_dir = os.getenv("QCF_BASE_FOLDER")
    elif config_data.get("base_folder") is not None:
        base_dir = config_data["base_folder"]
    elif len(file_paths) > 0:
        # use the location of the last file as the base directory
        base_dir = os.path.dirname(file_paths[-1])
    else:
        raise RuntimeError("Base folder must be specified somehow. Maybe set QCF_BASE_FOLDER in the environment?")

    config_data["base_folder"] = os.path.abspath(base_dir)

    # Test environment values for old names
    for key in os.environ.keys():
        if re.match(r"QCF_DB_[^_]", key):
            new_key = key.replace("QCF_DB_", "QCF_DATABASE__")
            raise RuntimeError(f"Environment variable {key} is deprecated. Use {new_key} instead.")
        if re.match(r"QCF_API_[^_]", key):
            new_key = key.replace("QCF_API_", "QCF_API__")
            raise RuntimeError(f"Environment variable {key} is deprecated. Use {new_key} instead.")
        if re.match("QCF_AUTORESET_[^_]", key):
            new_key = key.replace("QCF_AUTORESET_", "QCF_AUTO_RESET__")
            raise RuntimeError(f"Environment variable {key} is deprecated. Use {new_key} instead.")
        if re.match("QCF_S3_[^_]", key):
            new_key = key.replace("QCF_S3_", "QCF_S3__")
            raise RuntimeError(f"Environment variable {key} is deprecated. Use {new_key} instead.")

    # Pydantic will handle reading from environment variables
    # See if it can assemble a config. If there was a problem, and no
    # config files specified, mention that
    try:
        return FractalConfig(**config_data)
    except Exception as e:
        if len(file_paths) == 0:
            raise RuntimeError(f"Could not assemble a working configuration from environment variables:\n{str(e)}")
        raise


def write_initial_configuration(file_path: str, full_config: bool = True):
    base_folder = os.path.dirname(file_path)

    # Generate two secret keys for flask/jwt
    secret_key = secrets.token_urlsafe(32)
    jwt_secret_key = secrets.token_urlsafe(32)

    db_config = {
        "username": "qcfractal",
        "password": secrets.token_urlsafe(32),
    }

    default_config = FractalConfig(
        base_folder=base_folder, api={"secret_key": secret_key, "jwt_secret_key": jwt_secret_key}, database=db_config
    )

    default_config.database.port = find_open_port(starting_port=5432)
    default_config.api.port = find_open_port(starting_port=7777)

    include = None
    if not full_config:
        include = {
            "name": True,
            "enable_security": True,
            "log_access": True,
            "allow_unauthenticated_read": True,
            "logfile": True,
            "loglevel": True,
            "service_frequency": True,
            "max_active_services": True,
            "heartbeat_frequency": True,
            "database": {"own", "host", "port", "database_name", "base_folder", "username", "password"},
            "api": {"secret_key", "jwt_secret_key", "host", "port"},
        }

    with open(file_path, "x") as f:
        yaml.dump(default_config.dict(include=include), f, sort_keys=False)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcfractal.components.managers.db_models import ComputeManagerORM
from qcportal.exceptions import ComputeManagerError
from qcportal.managers import ManagerName, ManagerStatusEnum
from qcportal.utils import now_at_utc

if TYPE_CHECKING:
    from qcfractal.db_socket import SQLAlchemySocket
    from sqlalchemy.orm.session import Session


def test_manager_socket_deactivate_before_notasks(storage_socket: SQLAlchemySocket, session: Session):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    # UUID is different
    mname2 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5679")

    time_0 = now_at_utc()

    mid1 = storage_socket.managers.activate(
        name_data=mname1,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "qcprog": ["unknown"], "qcprog2": ["v3.0"]},
        compute_tags=["tag1", "tag2"],
    )

    time_1 = now_at_utc()
    mid2 = storage_socket.managers.activate(
        name_data=mname2,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "qcprog": ["unknown"], "qcprog2": ["v3.0"]},
        compute_tags=["tag1", "tag2"],
    )

    name1 = mname1.fullname
    name2 = mname2.fullname

    deactivated = storage_socket.managers.deactivate(modified_before=time_0)
    assert deactivated == []

    deactivated = storage_socket.managers.deactivate(modified_before=time_1)
    assert deactivated == [name1]

    manager1 = session.get(ComputeManagerORM, mid1)
    manager2 = session.get(ComputeManagerORM, mid2)
    assert manager1.status == ManagerStatusEnum.inactive
    assert manager2.status == ManagerStatusEnum.active

    deactivated = storage_socket.managers.deactivate(modified_before=now_at_utc())
    assert deactivated == [name2]

    session.expire_all()
    manager1 = session.get(ComputeManagerORM, mid1)
    manager2 = session.get(ComputeManagerORM, mid2)
    assert manager1.status == ManagerStatusEnum.inactive
    assert manager2.status == ManagerStatusEnum.inactive


def test_manager_socket_get_proj(storage_socket: SQLAlchemySocket):
    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    # UUID is different
    mname2 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5679")

    storage_socket.managers.activate(
        mname1,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "qcprog": ["unknown"], "qcprog2": ["v3.0"]},
        compute_tags=["tag1", "tag2"],
    )

    storage_socket.managers.activate(
        name_data=mname2,
        manager_version="v2.0",
        username="bill",
        programs={"qcengine": ["unknown"], "qcprog": ["unknown"], "qcprog2": ["v3.0"]},
        compute_tags=["tag1", "tag2"],
    )

    name1 = mname1.fullname
    name2 = mname2.fullname

    # Logs not included by default
    manager = storage_socket.managers.get([name1, name2])
    assert "log" not in manager[0]
    assert "log" not in manager[1]

    manager = storage_socket.managers.get([name1, name2], include=["cluster"])
    assert manager[0]["name"] == name1
    assert manager[1]["name"] == name2
    assert set(manager[0].keys()) == {"id", "name", "cluster"}  # name & id always returned
    assert set(manager[1].keys()) == {"id", "name", "cluster"}

    manager = storage_socket.managers.get([name1, name2], include=["name"])
    assert set(manager[0].keys()) == {"id", "name"}
    assert set(manager[1].keys()) == {"id", "name"}

    # cannot exclude name
    manager = storage_socket.managers.get([name1, name2], exclude=["cluster", "name"])
    assert "name" in manager[0]
    assert "name" in manager[1]
    assert "cluster" not in manager[0]
    assert "cluster" not in manager[1]


def test_manager_socket_update_resource_stats_nonexist(storage_socket: SQLAlchemySocket):
    with pytest.raises(ComputeManagerError, match=r"does not exist"):
        storage_socket.managers.update_resource_stats(
            "no_manager_name",
            total_cpu_hours=1.234,
            active_tasks=3,
            active_cores=10,
            active_memory=3.45,
        )



import base64
import json
from hashlib import sha256
from typing import Any

import numpy as np

try:
    from pydantic.v1 import BaseModel
    from pydantic.v1.json import pydantic_encoder
except ImportError:
    from pydantic import BaseModel
    from pydantic.json import pydantic_encoder


class _JSONEncoder_1(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        # JSON does not handle byte arrays
        # So convert to base64
        if isinstance(obj, bytes):
            return {"_bytes_base64_": base64.b64encode(obj).decode("ascii")}

        # Now do anything with pydantic, excluding unset fields
        # Also always use aliases when serializing
        if isinstance(obj, BaseModel):
            return obj.dict(exclude_unset=True, by_alias=True)

        # Let pydantic handle other things
        try:
            return pydantic_encoder(obj)
        except TypeError:
            pass

        # Flatten numpy arrays
        # This is mostly for Molecule class
        # TODO - remove once all data in the database in converted
        if isinstance(obj, np.ndarray):
            if obj.shape:
                return obj.ravel().tolist()
            else:
                return obj.tolist()

        return json.JSONEncoder.default(self, obj)


def hash_dict_1(d):
    j = json.dumps(d, ensure_ascii=True, sort_keys=True, cls=_JSONEncoder_1).encode("utf-8")
    return sha256(j).hexdigest()



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from qcportal.exceptions import InvalidArgumentsError, MissingDataError

if TYPE_CHECKING:
    from qcarchivetesting.testing_classes import QCATestingSnowflake


def test_singlepoint_dataset_socket_entries_from_errors(snowflake: QCATestingSnowflake):
    snowflake_client = snowflake.client()
    src_opt_ds = snowflake_client.add_dataset("optimization", "Test src optimization dataset")
    src_sp_ds = snowflake_client.add_dataset("singlepoint", "Test src singlepoint dataset")

    sp_ds = snowflake_client.add_dataset("singlepoint", "Test singlepoint dataset")

    socket = snowflake.get_storage_socket()

    # Note - src_opt_ds.id + src_ds.id + sp_ds.id + 1 guaranteed not to exist
    missing_ds_id = src_opt_ds.id + src_sp_ds.id + sp_ds.id + 1

    # Destination dataset id does not exist
    with pytest.raises(MissingDataError, match="Cannot find dataset"):
        socket.datasets.singlepoint.add_entries_from_ds(
            dataset_id=missing_ds_id,
            from_dataset_id=src_opt_ds.id,
            from_dataset_type=None,
            from_dataset_name=None,
            from_specification_name="test_spec",
        )

    # Source dataset id does not exist
    with pytest.raises(MissingDataError, match="Cannot find dataset"):
        socket.datasets.singlepoint.add_entries_from_ds(
            dataset_id=sp_ds.id,
            from_dataset_id=missing_ds_id,
            from_dataset_type=None,
            from_dataset_name=None,
            from_specification_name="test_spec",
        )
    # Dataset type/name does not exist
    with pytest.raises(MissingDataError, match="Cannot find dataset"):
        socket.datasets.singlepoint.add_entries_from_ds(
            dataset_id=sp_ds.id,
            from_dataset_id=None,
            from_dataset_type="optimization",
            from_dataset_name="Does not exist",
            from_specification_name="test_spec",
        )

    # Bad dataset type for the given id
    with pytest.raises(InvalidArgumentsError, match="not singlepoint"):
        socket.datasets.singlepoint.add_entries_from_ds(
            dataset_id=sp_ds.id,
            from_dataset_id=src_opt_ds.id,
            from_dataset_type="singlepoint",
            from_dataset_name=None,
            from_specification_name="test_spec",
        )

    # Bad dataset type for the given id
    with pytest.raises(InvalidArgumentsError, match="not optimization"):
        socket.datasets.singlepoint.add_entries_from_ds(
            dataset_id=sp_ds.id,
            from_dataset_id=src_sp_ds.id,
            from_dataset_type="optimization",
            from_dataset_name=None,
            from_specification_name="test_spec",
        )

    # Optimization ds, but no specification
    with pytest.raises(InvalidArgumentsError, match="from_specification_name must be provided"):
        socket.datasets.singlepoint.add_entries_from_ds(
            dataset_id=sp_ds.id,
            from_dataset_id=src_opt_ds.id,
            from_dataset_type=None,
            from_dataset_name=None,
            from_specification_name=None,
        )



from __future__ import annotations

import hashlib
import importlib
import logging
import os
import uuid
from typing import TYPE_CHECKING

from sqlalchemy import select

from qcportal.exceptions import MissingDataError
from qcportal.external_files import ExternalFileTypeEnum, ExternalFileStatusEnum
from .db_models import ExternalFileORM

# Boto3 package is optional
_boto3_spec = importlib.util.find_spec("boto3")

if _boto3_spec is not None:
    boto3 = importlib.util.module_from_spec(_boto3_spec)
    _boto3_spec.loader.exec_module(boto3)


if TYPE_CHECKING:
    from sqlalchemy.orm.session import Session
    from qcfractal.components.internal_jobs.status import JobProgress
    from qcfractal.db_socket.socket import SQLAlchemySocket
    from typing import Optional, Dict, Any, Tuple, Union, BinaryIO, Callable, Generator


class ExternalFileSocket:
    """
    Socket for managing/querying external files
    """

    def __init__(self, root_socket: SQLAlchemySocket):
        self.root_socket = root_socket
        self._logger = logging.getLogger(__name__)
        self._s3_config = root_socket.qcf_config.s3

        if not self._s3_config.enabled:
            self._logger.info("S3 service for external files is not configured")
            return
        elif _boto3_spec is None:
            raise RuntimeError("boto3 package is required for S3 support")

        self._s3_client = boto3.client(
            "s3",
            endpoint_url=self._s3_config.endpoint_url,
            aws_access_key_id=self._s3_config.access_key_id,
            aws_secret_access_key=self._s3_config.secret_access_key,
            verify=self._s3_config.verify,
        )

        # may raise an exception (bad/missing credentials, etc)
        server_bucket_info = self._s3_client.list_buckets()
        server_buckets = {k["Name"] for k in server_bucket_info["Buckets"]}
        self._logger.info(f"Found {len(server_buckets)} buckets on the S3 server/account")

        # Make sure the buckets we use exist
        self._bucket_map = self._s3_config.bucket_map
        if self._bucket_map.dataset_attachment not in server_buckets:
            raise RuntimeError(f"Bucket {self._bucket_map.dataset_attachment} (for dataset attachments)")

    def _lookup_bucket(self, file_type: Union[ExternalFileTypeEnum, str]) -> str:
        # Can sometimes be a string from sqlalchemy (if set because of polymorphic identity)
        if isinstance(file_type, str):
            return getattr(self._bucket_map, file_type)
        elif isinstance(file_type, ExternalFileTypeEnum):
            return getattr(self._bucket_map, file_type.value)
        else:
            raise ValueError("Unknown parameter type for lookup_bucket: ", type(file_type))

    def add_data(
        self,
        file_data: BinaryIO,
        file_orm: ExternalFileORM,
        *,
        job_progress: Optional[JobProgress] = None,
        session: Optional[Session] = None,
    ) -> int:
        """
        Add raw file data to the database.

        This will fill in the appropriate fields on the given ORM object.
        The file_name and file_type must be filled in already.

        In this function, the `file_orm` will be added to the session and the session will be flushed.
        This is done at the beginning to show that the file is "processing", and at the end when the
        addition is completed.

        Parameters
        ----------
        file_data
            Binary data to be read from
        file_orm
            Existing ORM object that will be filled in with metadata
        job_progress
            Object used to track progress if this function is being run in a background job
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            ID of the external file (which is also set in the given ORM object)
        """

        bucket = self._lookup_bucket(file_orm.file_type)

        self._logger.info(
            f"Uploading data to S3 bucket {bucket}. file_name={file_orm.file_name} type={file_orm.file_type}"
        )
        object_key = str(uuid.uuid4())
        sha256 = hashlib.sha256()
        file_size = 0

        multipart_upload = self._s3_client.create_multipart_upload(Bucket=bucket, Key=object_key)
        upload_id = multipart_upload["UploadId"]
        parts = []
        part_number = 1

        with self.root_socket.optional_session(session) as session:
            file_orm.status = ExternalFileStatusEnum.processing
            file_orm.bucket = bucket
            file_orm.object_key = object_key
            file_orm.sha256sum = ""
            file_orm.file_size = 0

            session.add(file_orm)
            session.flush()

            try:
                # Chunk size = 256MiB. There is a limit of 10,000 chunks, which would be
                # 2.56TiB
                while chunk := file_data.read(256 * 1024 * 1024):
                    if job_progress is not None:
                        job_progress.raise_if_cancelled()

                    sha256.update(chunk)
                    file_size += len(chunk)

                    response = self._s3_client.upload_part(
                        Bucket=bucket, Key=object_key, PartNumber=part_number, UploadId=upload_id, Body=chunk
                    )
                    parts.append({"PartNumber": part_number, "ETag": response["ETag"]})
                    part_number += 1

                self._s3_client.complete_multipart_upload(
                    Bucket=bucket, Key=object_key, UploadId=upload_id, MultipartUpload={"Parts": parts}
                )

            except Exception as e:
                self._s3_client.abort_multipart_upload(Bucket=bucket, Key=object_key, UploadId=upload_id)
                raise e

            self._logger.info(f"Uploading data to S3 bucket complete. Finishing writing metadata to db")
            file_orm.status = ExternalFileStatusEnum.available
            file_orm.sha256sum = sha256.hexdigest()
            file_orm.file_size = file_size
            session.flush()

            return file_orm.id

    def add_file(
        self,
        file_path: str,
        file_orm: ExternalFileORM,
        *,
        job_progress: Optional[JobProgress] = None,
        session: Optional[Session] = None,
    ) -> int:
        """
        Add an existing file to the database

        See documentation for :ref:`add_data` for more information.

        Parameters
        ----------
        file_path
            Path to an existing file to be read from
        file_orm
            Existing ORM object that will be filled in with metadata
        job_progress
            Object used to track progress if this function is being run in a background job
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            ID of the external file (which is also set in the given ORM object)
        """

        file_size = os.path.getsize(file_path)
        self._logger.info(f"Uploading {file_path} to S3. File size: {file_size/1048576} MiB")

        with open(file_path, "rb") as f:
            return self.add_data(f, file_orm, job_progress=job_progress, session=session)

    def get_metadata(
        self,
        file_id: int,
        *,
        session: Optional[Session] = None,
    ) -> Dict[str, Any]:
        """
        Obtain external file information

        Parameters
        ----------
        file_id
            ID for the external file
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            List of molecule data (as dictionaries) in the same order as the given ids.
            If missing_ok is True, then this list will contain None where the molecule was missing.
        """

        with self.root_socket.optional_session(session, True) as session:
            ef = session.get(ExternalFileORM, file_id)
            if ef is None:
                raise MissingDataError(f"Cannot find external file with id {file_id} in the database")

            return ef.model_dict()

    def delete(self, file_id: int, *, session: Optional[Session] = None):
        """
        Deletes an external file from the database and from remote storage

        Parameters
        ----------
        file_id
            ID of the external file to remove
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            Metadata about what was deleted and any errors that occurred
        """

        with self.root_socket.optional_session(session) as session:
            stmt = select(ExternalFileORM).where(ExternalFileORM.id == file_id)
            ef_orm = session.execute(stmt).scalar_one_or_none()

            if ef_orm is None:
                raise MissingDataError(f"Cannot find external file with id {file_id} in the database")

            bucket = ef_orm.bucket
            object_key = ef_orm.object_key

            session.delete(ef_orm)
            session.flush()

            # now delete from S3 storage
            self._s3_client.delete_object(Bucket=bucket, Key=object_key)

    def get_url(self, file_id: int, *, session: Optional[Session] = None) -> Tuple[str, str]:
        """
        Obtain an url that a user can use to download the file directly from the S3 bucket

        Will raise an exception if the file_id does not exist

        Parameters
        ----------
        file_id
            ID of the external file
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            File name and direct URL to the file
        """

        with self.root_socket.optional_session(session, True) as session:
            ef = session.get(ExternalFileORM, file_id)
            if ef is None:
                raise MissingDataError(f"Cannot find external file with id {file_id} in the database")

            url = self._s3_client.generate_presigned_url(
                ClientMethod="get_object",
                Params={
                    "Bucket": ef.bucket,
                    "Key": ef.object_key,
                    "ResponseContentDisposition": f'attachment; filename = "{ef.file_name}"',
                },
                HttpMethod="GET",
                ExpiresIn=120,
            )

            return ef.file_name, url

    def get_file_streamer(
        self, file_id: int, *, session: Optional[Session] = None
    ) -> Tuple[str, Callable[[], Generator[bytes, None, None]]]:
        """
        Returns a function that streams a file from an S3 bucket.

        Parameters
        ----------
        file_id
            ID of the external file
        session
            An existing SQLAlchemy session to use. If None, one will be created. If an existing session
            is used, it will be flushed (but not committed) before returning from this function.

        Returns
        -------
        :
            The recommended filename and a generator function that yields chunks of the file
        """

        with self.root_socket.optional_session(session, True) as session:
            ef = session.get(ExternalFileORM, file_id)
            if ef is None:
                raise MissingDataError(f"Cannot find external file with id {file_id} in the database")

            s3_obj = self._s3_client.get_object(Bucket=ef.bucket, Key=ef.object_key)

            def _generator_func():
                for chunk in s3_obj["Body"].iter_chunks(chunk_size=(8 * 1048576)):
                    yield chunk

            return ef.file_name, _generator_func



# This file is used in the flask_app and db stuff to register routes
# (ie, this file is imported, which causes all the routes
# to be registered with the blueprint)

from .molecules import db_models, routes
from .auth import db_models, routes
from .serverinfo import db_models, routes
from .managers import db_models, routes
from .tasks import db_models, routes
from .services import db_models
from .internal_jobs import db_models, routes
from .external_files import db_models, routes

from . import record_db_models, dataset_db_models, project_db_models, record_routes, dataset_routes, project_routes
from .singlepoint import record_db_models, dataset_db_models, routes
from .optimization import record_db_models, dataset_db_models, routes
from .torsiondrive import record_db_models, dataset_db_models, routes
from .gridoptimization import record_db_models, dataset_db_models, routes
from .reaction import record_db_models, dataset_db_models, routes
from .manybody import record_db_models, dataset_db_models, routes
from .neb import record_db_models, dataset_db_models, routes

from . import record_db_views, dataset_db_views



from __future__ import annotations

import pytest

from qcarchivetesting.testing_classes import QCATestingSnowflake
from qcportal import PortalRequestError
from qcportal.managers import ManagerName, ManagerStatusEnum
from qcportal.utils import now_at_utc


def test_manager_client_get(snowflake: QCATestingSnowflake):
    time_0 = now_at_utc()

    mname1 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5678")

    # UUID is different
    mname2 = ManagerName(cluster="test_cluster", hostname="a_host", uuid="1234-5678-1234-5679")

    client = snowflake.client()

    mclient1 = snowflake.manager_client(mname1, username="bill")
    mclient2 = snowflake.manager_client(mname2, username="bill")

    time_0 = now_at_utc()
    mclient1.activate(
        manager_version="v2.0",
        programs={"qcengine": ["unknown"], "qcprog": ["unknown"], "qcprog2": ["v3.0"]},
        compute_tags=["tag1", "tag2"],
    )

    time_1 = now_at_utc()
    mclient2.activate(
        manager_version="v2.0",
        programs={"qcengine": ["unknown"], "qcprog": ["unknown"], "qcprog2": ["v3.0"]},
        compute_tags=["tag1"],
    )
    time_2 = now_at_utc()

    name1 = mname1.fullname
    name2 = mname2.fullname

    # Test getting duplicates
    manager = client.get_managers([name2, name1, name1, name2])
    assert len(manager) == 4
    assert manager[1].name == name1
    assert manager[1].cluster == "test_cluster"
    assert manager[1].hostname == "a_host"
    assert manager[1].username == "bill"
    assert manager[1].compute_tags == ["tag1", "tag2"]
    assert manager[1].status == ManagerStatusEnum.active
    assert manager[1].created_on > time_0
    assert manager[1].modified_on > time_0
    assert manager[1].created_on < time_1
    assert manager[1].modified_on < time_1

    assert manager[0].name == name2
    assert manager[0].compute_tags == ["tag1"]
    assert manager[0].status == ManagerStatusEnum.active
    assert manager[0].created_on > time_1
    assert manager[0].modified_on > time_1
    assert manager[0].created_on < time_2
    assert manager[0].modified_on < time_2

    assert manager[2].id == manager[1].id
    assert manager[3].id == manager[0].id


def test_manager_client_get_nonexist(snowflake: QCATestingSnowflake):
    activated_manager_name, _ = snowflake.activate_manager()
    client = snowflake.client()
    manager = client.get_managers(["noname", activated_manager_name.fullname], missing_ok=True)
    assert manager[0] is None
    assert manager[1] is not None

    with pytest.raises(PortalRequestError):
        client.get_managers(["noname", activated_manager_name.fullname], missing_ok=False)


def test_manager_client_get_empty(snowflake: QCATestingSnowflake, activated_manager_name: ManagerName):
    # include activated_manager_name so that there is something in the db
    client = snowflake.client()
    manager = client.get_managers([], missing_ok=True)
    assert manager == []



"""Fix unique index on dataset table

Revision ID: 34d57d259c11
Revises: 6b24c66979ab
Create Date: 2024-04-07 09:36:05.692066

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = "34d57d259c11"
down_revision = "6b24c66979ab"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("ALTER TABLE base_dataset DROP CONSTRAINT IF EXISTS ux_base_dataset_dataset_type_lname")
    op.create_unique_constraint("ux_base_dataset_dataset_type_lname", "base_dataset", ["dataset_type", "lname"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



from qcfractal.components.auth.role_permissions import GLOBAL_ROLE_PERMISSIONS


def evaluate_global_permissions(role: str, resource: str, action: str):
    # Always paranoid about auth stuff
    if not isinstance(role, str) or len(role) == 0:
        raise TypeError("The role must be a string with length > 0")
    if not isinstance(resource, str) or len(resource) == 0:
        raise TypeError("The resource must be a string with length > 0")
    if not isinstance(action, str) or len(action) == 0:
        raise TypeError("The action must be a string with length > 0")

    if role not in GLOBAL_ROLE_PERMISSIONS:
        return False

    role_permissions = GLOBAL_ROLE_PERMISSIONS[role]

    if resource in role_permissions:
        resource_permissions = role_permissions[resource]
    elif "*" in role_permissions:
        resource_permissions = role_permissions["*"]
    else:
        return False

    if action in resource_permissions or "*" in resource_permissions:
        return True

    return False
